package numarea;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import numarea.GridField.MODE;

enum GAstage {	Pre, Select, Cross, Mutate	}

public class GA {
	private GridField 		gridfield;
	private	DifficultyPanel	difficultypanel;
	private	int				width;
	private	int				height;
	private	boolean[][]		pattern;
	private	boolean[][]		preshapepattern;
	private	int[][]			preshapeweight;
	
	private	int[][]			prenumtbl;

	private	int[][]			pegids;	// pegにIDを振る。0～、　-1は非peg。
	private	int				peglength;	
	private	ArrayList<Point>	pegs;
	
	// セルをカバーするpegを管理する。済みはnull。
	private	ArrayList<Point>[][]	coverpegs;	
	
	private	ArrayList<Individual>	indivs;		// 個体プール（現）
	private	ArrayList<Individual>	nexts;		// 個体プール（次）
	
	private	Cutting 	cutting = null;
	
	private	GAstage		stage;

	private	long		starttime;
	
	private	boolean		isrecdivyon;
	private	int			recdivlimit;
	private	int			recdivtry;
	
	private	double		selectrate;
	private	double		agingfactor;
	private	double		parentcurve;
	private	double		parentcut;
	private	int			maxmutation;
	private	int			maxgeneration;

	private	int			lowerlimit;
	private	int			upperlimit	;
	private	double		limitweight;
	
	private	boolean	isuneven;
	private	double		uneventhreshold;	// ≧
	private	int			unevenstep;
	private	int			unevenmaxcount;
	private	double		unevenweight;
	
	private	double		rectweight;
	
	private	double		restcountweight;
	private	double		mdivcountweight;
	private	double		prenumweight;
	private	double		prefixweight;
	private	int			uniquedistance;
	
	private	boolean	isrectprint, isunevenprint, isnumprint, ishidprint, 
									isrngprint, istimeprint;
	
	private	boolean	nomoregeneration;
	
	public		double		getuneventhreshold() {
		readgacount();
		return	uneventhreshold;
	}
	
	public		int		getunevenstep() {
		readgacount();
		return	unevenstep;
	}	
	
	//------------------------------------------------------------------------------
	//		Constructor	GA
	//------------------------------------------------------------------------------
	public GA ( GridField gf ) {
		gridfield = gf;
		difficultypanel = gridfield.difficultypanel;
		width = gridfield.getGridSize().width;
		height = gridfield.getGridSize().height;
		pattern = gridfield.getPattern();
		setupPegs();
		setupCoverPegs();
	}

	private	void	setupPegs() {
		pegs = new ArrayList<Point>();
		pegids = new int[width][height];
		int	id = 0;
		for( int x=0; x<width; ++x )
			for( int y=0; y<height; ++ y ) {
				if( pattern[x][y] ) {
					pegs.add(new Point(x,y));
					pegids[x][y] = id++;
				} else {
					pegids[x][y] = -1;
				}
			} 

		peglength = id;
	}
	
	private	void	setupCoverPegs() {
		coverpegs = (ArrayList<Point>[][])new ArrayList[width][height];
		for( int x=0; x<width; ++x )
			for( int y=0; y<height; ++ y ) {
				if( pegids[x][y] >= 0 ) {
					coverpegs[x][y] = null;
				} else {
					Point current = new Point(x,y);
					coverpegs[x][y] = new ArrayList<Point>();

					for( Point peg : pegs ) {
						if( iscoverablePeg( current, peg )) {
							coverpegs[x][y].add(peg);
						}
					}
				}
			}
	}
	
	private boolean	iscoverablePeg( Point cur, Point peg ) {
		int		x0 = Math.min( cur.x, peg.x );
		int		x1 = Math.max( cur.x, peg.x );
		int		y0 = Math.min( cur.y, peg.y );
		int		y1 = Math.max( cur.y, peg.y );
		Rectangle	rec = new Rectangle( x0, y0, x1-x0+1, y1-y0+1 );
		
		for( Point p : pegs ) {
			if( rec.contains(p) ) {
				if( ! p.equals(peg) ) {
					return	false;
				}
			}
		}
		return	true;
	}
	
	//	選択・交叉・突然変異のデフォルト値
	public		int		selectcount = 10;	
	public		int		crosscount  = 18;
	public		int		mutatecount =  2;
	
	private	int		realmutatecount;
	private	int		badloopcount = 0;
	
	private	int		generation;
	
	int	cuttrycount;
	
	private	int		minselect;
	private	int		maxselect;
	
	private	int		totalspaces;
	
	private Individual	individual;			// 成功個体
	
	public	void	initgacount() {
		selectcount = 10;
		crosscount  = 18;
		mutatecount = 2;
	}
	
	private	void	readgacount() {
		MakePanel mp = gridfield.makepanel;
		DifficultyPanel	dp = gridfield.difficultypanel;
		DesignPanel		dsp = gridfield.designpanel;
		
		isrecdivyon		= dsp.isrecdivyon();
		recdivlimit		= dsp.getrecdivlimit();
		recdivtry		= dsp.getrecdivtry();
		restcountweight	= dsp.getrestcountweight();
		mdivcountweight	= dsp.getmdivcountweight();
		prenumweight	= dsp.getprenumweight();
		prefixweight 	= dsp.getprefixweight();
		
		lowerlimit		= dp.getlowerlimit();
		upperlimit		= dp.getupperlimit();
		limitweight		= dp.getlimitweight();
		
		isuneven		= dp.isuneven();
		uneventhreshold	= dp.getuneventhreashold();
		unevenstep		= dp.getunevenstep();
		unevenmaxcount 	= dp.getunevenmaxcount();
		unevenweight	= dp.getunevenweight();
		rectweight		= dp.getrectweight();
		
		selectcount 	= mp.getselectcount();
		crosscount  	= mp.getcrosscount();
		mutatecount 	= mp.getmutatecount();
		selectrate		= mp.getselectrate();
		agingfactor		= mp.getagingfactor();
		parentcurve		= mp.getparentcurve();
		parentcut		= mp.getparentcut();
		maxmutation		= mp.getmaxmutation(); 
		maxgeneration	= mp.getmaxgeneration();
		uniquedistance	= mp.getuniquedistance();
		
		isrectprint		= mp.isrectprint();
		isunevenprint	= mp.isunevenprint();
		isnumprint		= mp.isnumprint();
		ishidprint		= mp.ishidprint();
		isrngprint		= mp.isrngprint();
		istimeprint		= mp.istimeprint();
	}
	
	//-----------------------------------------------------------------------
	//	実行
	//-----------------------------------------------------------------------
	public void	exec() {
		gridfield.setskippaint(true);

		execMain();
		
		gridfield.setskippaint(false);

		if( individual!=null ) {
			addGeneInfo(0,0,0);
		}
	}
	
	private	void	execMain() {
		gridfield.makepanel.clearprepost();
		gridfield.setBreak(false);
		gridfield.generating	= true;
		gridfield.genrecs = null;
		totalspaces = width * height;

		geneinfo = new ArrayList<GeneticPoolInfo>();
		
		individual = null;
		nomoregeneration = false;
		
		starttime = System.currentTimeMillis();
		updateexectime();
		clearlog();
		
		readgacount();
		
		realmutatecount = mutatecount;
		
		boolean	found = false;
		indivs = new ArrayList<Individual>();
		nexts = new ArrayList<Individual>();
		generation = 0;
		badloopcount = 0;
		
		gridfield.setMode(MODE.CUTTING,true);
		
		preshapepattern = gridfield.getPreShapeTable();
		if( !isactivepreshape() )
			preshapepattern = null;
		preshapeweight  = gridfield.getPreShapeWeight();
		
		prenumtbl = gridfield.getPreNumTable();
		if( !isactiveprenum() )
			prenumtbl = null;
		
		// 初期処理
		stage = GAstage.Pre;
		for( int i=0; i<selectcount+crosscount+mutatecount; ) {
			long	starttime = System.currentTimeMillis();
			long	current;

			Individual	indv = null;
			
			for(cuttrycount=0;;++cuttrycount) {
				if( gridfield.isBreak() )
					return;
				
				gridfield.clearRect();
				gridfield.clearAllArea();

				ArrayList<Rectangle> recs = freecut();
				if( recs == null )
					continue;				

				gridfield.setAreas( recs );
				
				indv = new Individual(recs, 'I' );
				indv.mdivcount = countMultiDiv(recs);
				
				if( check(indv.sequence) )
					break;
			}
			
			gridfield.setMode(MODE.SOLVE);
			
			shortsleep();
			
			gridfield.setMode(MODE.CUTTING);
			if( sameInPool( indv, indivs ) ) 
				continue;
			
			indv.restcount = gridfield.getBlankCount();
			indv.age = generation;
			indv.pushRecsonGrid();
			indv.rectshapeeval = difficultypanel.shapeeval( indv.getRectList() );
			indv.prenumeval = prefixnumbereval( indv.getRectList() );
			indv.prefixeval = prefixshapeeval( indv.getRectList() );
		
			gridfield.setMode(MODE.EVAL);	// 解き直し
			gridfield.forcedrefresh();
			shortsleep();

			indv.orgpoint = gridfield.solver.getDiffPoint();
			indv.point = estimate( indv.restcount, indv.orgpoint );
			indv.updatecmprange();
			indv.uneveneval = getexcessuneven();
			indv.reeval();
			
			difficultypanel.setCurrentPointLabel(indv.point);
			current = System.currentTimeMillis();
			indv.time = (current - starttime) / 1000.0;
			
			indivs.add(indv);
			
			updateexectime();
			String str = String.format("%2dI   %s%n", i, indv.toString());
			appendlog(str);
			if( indv.eval()==0 /*&& indv.cmprange == 0*/ ) {
				individual = indv;
				break;
			} 
			++i;
		}
		
		appendlog(String.format("%n"));
		
		while( !found && !nomoregeneration ) {
			readgacount();
			
			for( Individual indv : indivs ) {
				indv.reeval();
			}
			
			Collections.sort(indivs);
			
			++generation;
							
			//	選択
			stage = GAstage.Select;
			int		evaluation;
			if( selectcount > 0 ) {
				nexts = getSelection( indivs, selectcount );		// 選択処理、間引きあり
				int	i=0;
				for( Individual id : nexts ) {
					if( id.eval()==0 )
						found = true;
					updateexectime();
					String str = String.format("%2dS   %s%n", i++, id.toString() );
					appendlog(str);
				}
				if( found ) {
					indivs = nexts;
					break;
				}

				evaluation = getEvalSum( indivs, selectcount );

				minselect = maxselect = nexts.get(0).eval();
				double	sumselect = 0;
				for( Individual sid : nexts ) {
					sumselect += sid.eval();
					int	v = sid.eval();
					if( minselect > v )
						minselect = v;
					if( maxselect < v )
						maxselect = v;
				}
				addGeneInfo(minselect,maxselect, sumselect/nexts.size());

			} else {
				int	evcount = crosscount/2;
				evaluation = getEvalSum( indivs, evcount );
				minselect = maxselect = indivs.get(0).eval();
				double	sumselect = 0;
				for( int i=0; i< evcount; ++i ) {
					Individual sid = indivs.get(i);
					sumselect += sid.eval();
					int	v = sid.eval();
					if( minselect > v )
						minselect = v;
					if( maxselect < v )
						maxselect = v;					
				}
				nexts = new ArrayList<Individual>();
				addGeneInfo(minselect,maxselect, sumselect/evcount);
			}

			updatemininfo();		// MakePanelのminiinfoの書き換え

			//			交叉
			stage = GAstage.Cross;
			if( found = crossProcessing() ) {
				indivs = nexts;
				break;
			}
			
			if( gridfield.isBreak() )
				return;
			
			// 突然変異
			stage = GAstage.Mutate;
			int	mutateloopcount = mutatecount + badloopcount;
			if( mutateloopcount > maxmutation )
				mutateloopcount = maxmutation;
			if( found = mutateProcessing(mutateloopcount) ) {
				indivs = nexts;
				break;
			}

			indivs = nexts;
			
			if( gridfield.isBreak() )
				return;
			
			Collections.sort( indivs );
			int		nextevaluation;
			if( selectcount > 0 ) {
				nextevaluation = getEvalSum( indivs, selectcount );
			} else {
				nextevaluation = getEvalSum( indivs, (crosscount+mutatecount)/2 );
			}
			if( evaluation > nextevaluation )	{	// 何とか進化している
				badloopcount = 0;
			} else {		
				++badloopcount;
			}
			
			String str = String.format("%d -> %d   %d/%d %n%n", evaluation, nextevaluation, badloopcount, generation );
			appendlog(str);			
			gridfield.makepanel.setprepost( evaluation, nextevaluation, badloopcount, generation );
			
			if( badloopcount == 0 )
				continue;
			
			if( generation >= maxgeneration )
				break;
		}
		
		if( individual != null ) {
			String str = String.format( "%nGeneration end%n%s%n", individual.toString() );
			appendlog(str);	
			individual.restore();
			difficultypanel.shapeeval( individual.getRectList() );
		} else {		// individual == null
			String str = String.format( "Generation end!%n" );
			appendlog(str);
		}
		
		gridfield.setMode(MODE.EVAL);
		gridfield.forcedrefresh();
		
		gridfield.generating	= false;
	}
	
	private	boolean	isactiveprenum() {
		for( int x=0; x<width; ++x )
			for( int y=0; y<height; ++y )
				if( prenumtbl[x][y] != 0 )
					return	true;
		return	false;		
	}
	
	private	boolean	isactivepreshape() {
		for( int x=0; x<width; ++x )
			for( int y=0; y<height; ++y )
				if( preshapepattern[x][y] )
					return	true;
		return	false;
	}
	
	private void	updatemininfo() {
		int		min = Integer.MAX_VALUE;
		int		max = 0;
		int		cnt = 0;
		
		for( Individual id : nexts ) {
			int	v = id.eval();
			if( v < min )
				min = v;
			if( v > max )
				max = v;
		}

		for( Individual id : nexts )
			if( id.eval() == min )
				++cnt;
		
		gridfield.makepanel.setselectinfo( min, cnt, max );
	}
	
	//------------------------------------------------------------------
	//	交叉処理
	boolean crossProcessing() {
		
		for( int k=0; k<crosscount; ++k ) {
			// ２組の個体を取り出す
			Individual indv = null;
			
			long	starttime = System.currentTimeMillis();
			
			for(;;) {
				if( gridfield.isBreak() )
					return false;

				int	poolsize = indivs.size();
				
				int	idx1, idx2;
				do {
					idx1 = nextSqInt(poolsize, parentcurve,parentcut);
					idx2 = nextSqInt(poolsize, parentcurve,parentcut);
				} while ( idx1 == idx2 );
				
				Individual indv1 = indivs.get(idx1);
				Individual indv2 = indivs.get(idx2);				
				indv = mix(indv1,indv2);
				
				if( indv == null )
					continue;
			
				if( sameInPool( indv, nexts ) )
					continue;
				
				if( check( indv.sequence ) )
					break;
			};
			long	current = System.currentTimeMillis();
			indv.time = (current - starttime) / 1000.0;
			nexts.add(indv);
			
			updateexectime();
			String str = String.format("%2dC %2s%s%n", k, indv.getInovMark(), indv.toString());
			appendlog(str);
			
			gridfield.refresh();
			
			if( indv.eval()==0 ) {
				individual = indv;
				return	true;
			}
		}
		return	false;
	}
	
	// 交叉生成
	Individual mix( Individual indv1, Individual indv2 ) {
		Rectangle[] seq1 = indv1.sequence;
		Rectangle[] seq2 = indv2.sequence;
		Rectangle[] newseq = new Rectangle[peglength];
		
		// 一致部分は、そのまま複製
		for( int i=0; i<peglength; ++i ) {
			Rectangle	r1 = seq1[i];
			Rectangle	r2 = seq2[i];
			if( r1==null || r2==null ) {	// 対処療法、良くない
				continue;
			}
			if( seq1[i].equals(seq2[i]) ) {
				newseq[i] = seq1[i];
			}
		}

		// ２つの領域に分ける
		Rectangle	bigrec = null;
		for( int i=0; i<peglength; ++i ) {
			if( newseq[i] != null )
				continue;
			if( bigrec == null ) {
				bigrec = seq1[i];
			} else {
				bigrec = bigrec.union(seq1[i]);
			}
		}
		if( bigrec == null )
			bigrec = new Rectangle(0,0,width,height);
		
		int	deltax = Util.rand.nextInt((bigrec.width+2)/2) + bigrec.width/4;
		int	deltay = Util.rand.nextInt((bigrec.height+2)/2)+ bigrec.height/4;		
		Point center = new Point( bigrec.x+deltax, bigrec.y+deltay);
		
		Rectangle big1 = null;
		Rectangle big2 = null;
		if( Util.rand.nextDouble() < 0.5 ) {	// 水平分割
			big1 = new Rectangle( 0, 0, width, center.y );
			big2 = new Rectangle( 0, center.y, width, height-center.y );		
		} else {									// 垂直分割
			big1 = new Rectangle( 0, 0, center.x, height );
			big2 = new Rectangle( center.x, 0, width-center.x, height );	
		}
		
		// seq1のbig1に完全に含まれるものをすべて追加
		// seq2のbig2に完全に含まれるものをすべて追加
		for( int i=0; i<peglength; ++i ) {
			if( newseq[i] != null )
				continue;
			if( big1.contains(seq1[i]) ) {
				newseq[i] = seq1[i];
			} else if( big2.contains(seq2[i]) ) {
				newseq[i] = seq2[i];
			}
		}
		
		cutting = new Cutting(width,height);
		cutting.precut(gridfield.getPatternTable(),newseq,false);
		
		ArrayList<Rectangle> rectangles = cutting.getRects();
		if( ! check(rectangles) )
			return	null;
		
		gridfield.setAreas( safeclone(rectangles) );				
		Individual indv = new Individual(rectangles, 'C' );

		gridfield.setMode(MODE.SOLVE);
		
		gridfield.setMode(MODE.CUTTING);
		indv.restcount = gridfield.getBlankCount();
		indv.mdivcount = countMultiDiv(rectangles);
		indv.age = generation;	
		indv.pushRecsonGrid();
		indv.rectshapeeval = difficultypanel.shapeeval(indv.getRectList());
		indv.prenumeval = prefixnumbereval( indv.getRectList() );
		indv.prefixeval = prefixshapeeval( indv.getRectList() );

		if( indv.restcount == 0 ) {
			gridfield.setMode(MODE.EVAL);	// 解き直し
			indv.point = gridfield.solver.getDiffPoint();
			indv.updatecmprange();
		} else {
			gridfield.setMode(MODE.EVAL);	// 解き直し
			indv.orgpoint = gridfield.solver.getDiffPoint();
			indv.point = estimate( indv.restcount, indv.orgpoint );
			indv.updatecmprange();
		}
		indv.uneveneval = getexcessuneven();
		indv.reeval();
		
		difficultypanel.setCurrentPointLabel(indv.point);
		gridfield.forcedrefresh();
		shortsleep();

		return	indv;
	}
	
	//------------------------------------------------------------------
	// 突然変異
	boolean	mutateProcessing( int maxloopcount ) {
		realmutatecount = 0;
		if( mutatecount == 0 )
			return	false;
		
		for( int k=0; k<maxloopcount; ) {
			if( gridfield.isBreak() )
				return false;
			
			Individual	indv;
			
			long	starttime = System.currentTimeMillis();
			
			for(;;) {
				// 個体を取り出す
				indv = indivs.get(nextSqInt(indivs.size(),parentcurve,parentcut));
				indv.restore();
				gridfield.setMode(MODE.SOLVE);
				shortsleep();

				gridfield.setMode(MODE.CUTTING);
				indv.restore();
				ArrayList<Rectangle> recs = null;

				Rectangle[]	recseq;

				if( indv.prefixeval > 0 && Util.rand.nextDouble()<0.8) {	// prefix条件に合わない
					Rectangle[] okrecs = prefixOKrecs(indv.sequence);
					recseq = removeSomeRecs( okrecs, indv.sequence, 1, true );

				} else {
					if( indv.restcount==0 && indv.rectshapeeval > 0 ) {
						// 形状のみ条件に合わない
						recs = gridfield.shapeerrorrecs(new ArrayList<Rectangle>(Arrays.asList(indv.sequence)));
					} else {
						recs = gridfield.getFixRectangles();	// 確定長方形リスト
					}
					recseq = convtoArray(recs);
					recseq = removeSomeRecs( recseq, indv.sequence, 1, false );		// 適当な個数の長方形を取り除く
				}

				cutting = new Cutting(width,height);
				cutting.precut(gridfield.getPatternTable(), recseq, true );

				ArrayList<Rectangle> rectangles = cutting.getRects();
				gridfield.setAreas( safeclone(rectangles) );			

				indv = new Individual( rectangles, 'M' );
				indv.mdivcount = countMultiDiv(rectangles);

				if( sameInPool( indv, nexts ) )
					continue;
				if( ! check( indv.sequence ) )
					continue;
				break;
			}
			
			gridfield.setMode(MODE.SOLVE);
			
			gridfield.setMode(MODE.CUTTING);
			indv.restcount = gridfield.getBlankCount();
			indv.age = generation;
			indv.pushRecsonGrid();
			indv.rectshapeeval = difficultypanel.shapeeval( indv.getRectList() );
			indv.prenumeval = prefixnumbereval( indv.getRectList() );
			indv.prefixeval = prefixshapeeval( indv.getRectList() );

			if( indv.restcount == 0 ) {
				gridfield.setMode(MODE.EVAL);	// 解き直し
				indv.point = gridfield.solver.getDiffPoint();
				indv.updatecmprange();
			} else {
				gridfield.setMode(MODE.EVAL);	// 解き直し
				indv.orgpoint = gridfield.solver.getDiffPoint();
				indv.point = estimate( indv.restcount, indv.orgpoint );
				indv.updatecmprange();
			}
			indv.uneveneval = getexcessuneven();
			indv.reeval();

			difficultypanel.setCurrentPointLabel(indv.point);
			long	current = System.currentTimeMillis();
			indv.time = ( current - starttime ) / 1000.0;
			
			gridfield.forcedrefresh();
			shortsleep();
			
			nexts.add(indv);

			updateexectime();
			String str = String.format("%2dM %2s%s%n", k, indv.getInovMark(), indv.toString() );
			appendlog(str);			
			
			++realmutatecount;
			
			gridfield.refresh();
			if( indv.eval()==0 && indv.cmprange == 0 ) {
				individual = indv;
				return	true;
			}
			
			++k;
		}
		return	false;
	}
	
	private	int		estimate( int rcnt, int val ) {
		if( rcnt == totalspaces || val == 0 ) {
			return	width * height * 10000;		// return big value;
		}
		int	rval = (int)((double)val * totalspaces / (totalspaces - rcnt));
		if( rval < 0 ) {
			System.out.printf("rcnt=%d, val=%d  -> %d%n", rcnt, val, rval );
		}
		
		return	rval;
	}
	
	private	Rectangle[]	convtoArray( ArrayList<Rectangle> recs ) {
		Rectangle[]	array = new Rectangle[peglength];
		
		for( int i=0; i<recs.size(); ++i ) {
			Rectangle rec = recs.get(i);
			int	idx = getId(rec);
			if( idx < 0 ) {	// ペグに属さない
				return	null;
			}
			array[idx] = rec;
		}
		
		return	array;
	}
	
	private	ArrayList<Rectangle>	safeclone( ArrayList<Rectangle> recs ) {
		ArrayList<Rectangle> list = new ArrayList<Rectangle>();
		for( Rectangle rec : recs ) {
			list.add(new Rectangle(rec));
		}
		return	list;
	}

	private	int		nextSqInt( int n, double pw, double cut ) {
		double	r = Util.rand.nextFloat();
		int    ni = (int) (n * cut * Math.exp( pw * Math.log(r)));
		return	ni;
	}
	
	//------------------------------------------------------------------
	// 選択
	private ArrayList<Individual> getSelection( ArrayList<Individual> indvs, int count ) {	
		ArrayList<Individual> next = new ArrayList<Individual>();
		ArrayList<Individual> srcs = new ArrayList<Individual>(indvs);
		
		for( int i=0; i<count; ) {
			Individual get = null;
			if( i==0 ) {			// 最初は間引かない
				get = srcs.get(0);
			} else {
				for( Individual id : srcs ) {
					if( Util.rand.nextDouble() < selectrate ) {		// 若干の間引き
						get = id;
						break;
					}	
				}
			}
			if( get == null )
				continue;

			srcs.remove(get);
			
			next.add(get);

			if( srcs.size() == 0 ) {
				break;
			}

			++i;
		}
		Collections.sort( next );
				
		return	next;
	}
	
	private int getEvalSum( ArrayList<Individual> indvs, int cnt ) {
		int	sum = 0;
		for( int i=0; i<cnt && i<indvs.size(); ++i ) {
			Individual indv = indvs.get(i);
			sum += indv.eval();
		}
		return	sum;
	}
	
	//------------------------------------------------------------------------------------------------------------
	//	初期の自由分割
	//------------------------------------------------------------------------------------------------------------
	private	ArrayList<Rectangle>	freecut() {
		boolean[][] backup	= Util.deepcopy(pattern);	// とりあえずの回避処置
		
		ArrayList<Rectangle> rectangles = freecutsub(0,0,width,height,gridfield.getPatternTable());
		
		pattern = backup;
		
		return	rectangles;
	}
	
	private	ArrayList<Rectangle>	freecut1( int x, int y, int w, int h, boolean pat[][] ) {
		for( int i=0; i<10; ++i ) {
			cutting = new Cutting(x,y,w,h);
			cutting.precut(pat,null,false);
			if( cutting.countNoBelong() == 0 ) {
				ArrayList<Rectangle> rectangles = cutting.getRects();
				return	rectangles;
			}
		}

		return	null;
	}
	
	private	ArrayList<Rectangle>		freecutsub( int x, int y, int w, int h, boolean pat[][] ) {
		int pegn = countpegs(w,h,pat);
				
		if( !isrecdivyon || pegn < recdivlimit ) {
			ArrayList<Rectangle>	list = freecut1(x,y,w,h,pat);
			if( list == null )
				return	null;
			ArrayList<Rectangle> dived = new ArrayList<Rectangle>();
			for( Rectangle r : list ) {	// 親座標系に変更
				Rectangle	mr = new Rectangle( r.x+x, r.y+y, r.width, r.height );
				dived.add(mr);
			}
			return	dived;
		}

		// 2分割する

		ArrayList<Rectangle>	recs = null;	

		for( int i=0; i<recdivtry; ++i ) {	// 再分割トライ回数
			double	r = (1.0+Util.rand.nextDouble())/3;
			ArrayList<Rectangle>	subprobs = new ArrayList<Rectangle>();
			if( w >= h ) {
				// 左右分割
				int	nw = (int)(w * r);
				Rectangle	wr = new Rectangle( x, y, nw, h );
				subprobs.add(wr);
				Rectangle  er = new Rectangle( x+nw, y, w-nw, h );
				subprobs.add(er);
			} else {
				int	nh = (int)(h * r);
				Rectangle nr = new Rectangle( x, y, w, nh );
				subprobs.add(nr);
				Rectangle sr = new Rectangle( x, y+nh, w, h-nh );
				subprobs.add(sr);	
			}

			recs = new ArrayList<Rectangle>();

			for( Rectangle subprob : subprobs ) {
				int offsetx = subprob.x;
				int offsety = subprob.y;
				int neww = subprob.width;
				int newh = subprob.height;
				boolean	newpat[][] = getsubpat( pat, subprob );
				ArrayList<Rectangle> cuts = null;
				cuts = freecutsub( offsetx, offsety, neww, newh, newpat );
				if( cuts == null ) {
					recs =	null;
					break;
				}

				for( Rectangle cr : cuts ) {
					recs.add(cr);
				}
			}
			
			// 正常に分割できている場合、分割結果を返す
			if( recs == null )
				continue;
			
			return	recs;
		}

		return	null;
	}
	
	private	int		countpegs(  int w, int h, boolean pat[][] ) {
		int	count = 0;
		
		for( int x=0; x<w; ++x )
			for( int y=0; y<h; ++y )
				if( pat[x][y] )
					++count;
		
		return	count;
	}
	
	private	boolean[][] 	getsubpat( boolean pat[][], Rectangle rec ) {
		boolean	newpat[][] = new boolean[rec.width][rec.height];
		boolean	orgpat[][] = gridfield.getPattern();
		for( int dx=0; dx<rec.width; ++dx )
			for( int dy=0; dy<rec.height; ++dy )
				newpat[dx][dy] = orgpat[rec.x + dx][rec.y + dy];
		return	newpat;
	}
	
	//------------------------------------------------------------------------------------------------------------	
	private	int	getId( Rectangle rec ) {
		for( int dx=0; dx<rec.width; ++dx )
			for( int dy=0; dy<rec.height; ++dy ) {
				int	id = pegids[rec.x+dx][rec.y+dy];
				if( id >= 0 )
					return	id;
			}

		return	-1;
	}

	boolean sameInPool( Individual indv, ArrayList<Individual>list ) {
		if( uniquedistance < 0 )
			return	false;
		
		boolean	same = false;
		
		for( Individual id : list ) {
			same = samecheck( indv, id, uniquedistance );
			if( same )
				break;
		}
		
		return	same;
	}
	
	private boolean	samecheck( Individual indv1, Individual indv2, int limit ) {
		if( limit < 0 )
			return	false;
		
		int	dist = 0;
		for( int i=0; i<indv1.sequence.length; ++i ) {
			if( ! indv1.sequence[i].equals( indv2.sequence[i] ) )
				++dist;
			if( dist > limit )
				return	false;
		}
		return	true;
	}
	
	// 長方形の並び　sequence　が条件を完全に満たしていること。
	boolean check( ArrayList<Rectangle> list ) {
		if( list == null || list.size() == 0 )
			return	false;
		Rectangle[]	recs = new Rectangle[list.size()];
		for( int i=0; i<list.size(); ++i )
			recs[i] = list.get(i);
		
		return	check(recs);		// 本当の検査は配列で行う
	}
	
	private boolean check( Rectangle[] sequence ) {
		if( sequence==null || sequence.length != peglength )
			return	false;
		
		boolean[][]	tbl = new boolean[width][height];

		int		sum = 0;
		
		for( int i=0; i<sequence.length; ++i ) {
			Rectangle rec = sequence[i];
			if( rec == null )
				return false;
			if( rec.width * rec.height <= 1 ) {	// サイズ１は不可
				return false;
			}
			if( ! rec.contains(pegs.get(i)) ) {	// 指定のペグを含む？
				return	false;
			}
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy ) {
					int u = rec.x + dx;
					int v = rec.y + dy;
					if( tbl[u][v] )
						return	false;
					tbl[u][v] = true;
				}
			sum += rec.width * rec.height;
		}
		
		if( sum != width * height )
			return	false;
		
		return	true;
	}
	
	//	未決定部分の長方形を少しだけ壊す。
	//		arecs	確定長方形
	//		qrecs	分割長方形（与問題）
	//		brcnt	破壊個数
	//		brwide	派手に壊す
	private Rectangle[]	removeSomeRecs( Rectangle[] arecs, Rectangle[] qrecs, 
												int brcnt, boolean brkwide ) {
		Rectangle[]	recs = qrecs.clone();
		
		// 破壊する長方形の候補を集め、サイズを拡大し、それと重なる長方形を全部消す
		Rectangle[]	brkareas = new Rectangle[brcnt];
		ArrayList<Rectangle>	candidates = new ArrayList<Rectangle>();
		for( int i=0; i<peglength; ++i ) {
			Rectangle ar = arecs[i];
			Rectangle qr = qrecs[i];
			if( ar!=null ) {
				if( ar.equals(qr) )
					continue;
			}	
			candidates.add(qr);
		}
		for( int j=0; j<brcnt; ) {
			if( candidates.size() == 0 )
				break;
			int	idx = Util.rand.nextInt(candidates.size());
			Rectangle brrec = candidates.get(idx);
			candidates.remove(idx);

			int		len = 1;

			if( Util.rand.nextInt(2) < 1)
				brkareas[j] = new Rectangle( brrec.x-len, brrec.y, brrec.width+2*len, brrec.height );
			else
				brkareas[j] = new Rectangle( brrec.x, brrec.y-len, brrec.width, brrec.height+2*len );
			++j;
		}
		for( int i=0; i<peglength; ++i ) {
			Rectangle rec = qrecs[i];
			boolean	overlap = false;
			for( int j=0; j<brcnt; ++j ) {
				Rectangle brec = brkareas[j];
				if( brec == null )
					continue;
				if( rec.intersects(brec) ) {
					overlap = true;
					break;
				}
			}
			if( overlap ) {
				Point p = pegs.get(i);
				recs[i] = new Rectangle( p.x, p.y, 1, 1 );
			}
		}
		return	recs;
	}

	private	void	shortsleep() {
		long	t = 100L + width*height/50;
		try {	Thread.sleep(t);	} catch (InterruptedException e) {}
	}
	
	//------------------------------------------------------------------------------
	//	局所的多重分割の箇所を数える
	//------------------------------------------------------------------------------
	private	int		countMultiDiv( ArrayList<Rectangle> rects ) {
		int		count = 0;
		
		count += count2x2MultiDiv( rects );
		
		return	count;
	}
	
	private 	int		count2x2MultiDiv( ArrayList<Rectangle> rects ) {
		int	count = 0;
		
		for( Rectangle rec1 : rects ) {
			if( rec1.width*rec1.height != 2 )
				continue;
			for( Rectangle rec2 : rects ) {
				if( rec2.width*rec2.height != 2 )
					continue;
				if( rec2.x<rec1.x )
					continue;
				if( rec2.x==rec1.x && rec2.y<rec1.y )
					continue;
				Rectangle rec = rec1.union(rec2);
				if( ! ( rec.width==2 && rec.height==2 ) )
					continue;

				if( (pattern[rec.x][rec.y] && pattern[rec.x+1][rec.y+1]) || 
							(pattern[rec.x][rec.y+1] && pattern[rec.x+1][rec.y]) ) {
					++count;
				}
			}
		}
		
		return	count;
	}
	
	//------------------------------------------------------------------------------
	//	Prefix number
	//------------------------------------------------------------------------------
	public	int		prefixnumbereval( ArrayList<Rectangle> reclist ) {
		int val = 0;
		if( reclist!=null && reclist.size() >0 ) {
			val = prenumbereval(reclist);
		}
		
		return	val;
	}
	
	public	int		prenumbereval( ArrayList<Rectangle> reclist ) {
		int	val = 0;
		if( prenumtbl == null )
			return	0;

		for( Rectangle rec : reclist ) {
			Point	p = whereisHint(rec);
			if( p == null )
				continue;
			int sz = prenumtbl[p.x][p.y];
			if( sz == 0 )
				continue;
			
			int rsize = rec.width * rec.height;
			if( sz == rsize )
				continue;

			val += Math.abs( sz - rsize );
		}		
	
		return	val;
	}
	
	//------------------------------------------------------------------------------
	//	Prefix shape
	//------------------------------------------------------------------------------
	public	int		prefixshapeeval( ArrayList<Rectangle> reclist ) {
		int	val = 0;
		if( reclist!=null && reclist.size() >0 ) {
			val = preshapeeval(reclist);
		}
		
		return	val;
	}
	
	public	int		preshapeeval( ArrayList<Rectangle> reclist ) {
		int	val = 0;
		if( preshapepattern == null )
			return	0;
		
		for( Rectangle rec : reclist ) {
			Point	peg = whereisHint(rec);
			boolean	isinside = preshapepattern[peg.x][peg.y];
			int	v = 0;
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy ) {
					if( isinside != preshapepattern[rec.x+dx][rec.y+dy] ) {
						v += preshapeweight[rec.x+dx][rec.y+dy];
						++v;
					}
				}
	
			val += v*v;
		}

		return	val;
	}
	
	// 突然変異のときに、残しておく長方形
	private	Rectangle[]	prefixOKrecs( Rectangle[] gene ) {
		if( preshapepattern == null )
			return	gene;
		
		Rectangle[] recs = new Rectangle[gene.length];
		
		for( int i=0; i<gene.length; ++i  ) {
			Rectangle	rec = gene[i];
			int	incount = 0;
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy )
					incount += preshapepattern[rec.x+dx][rec.y+dy] ? 1 : 0;
			int rsize = rec.width * rec.height;
			if( incount == 0 || rsize == incount )
				recs[i] = rec;
		}
		
		return	recs;
	}
	
	private Point whereisHint( Rectangle r ) {
		for( int i=0; i<r.width; ++i )
			for( int j=0; j<r.height; ++j )
				if( pattern[r.x+i][r.y+j] )
					return	new Point(r.x+i,r.y+j);
		return null;	
	}

	//------------------------------------------------------------------------------
	//	難易度テーブルの塊具合を求める
	//------------------------------------------------------------------------------
	private	int		getexcessuneven() {
		int		n = isuneven ? gridfield.getlumpcount( uneventhreshold, unevenstep, false ): 0;
		int		excess = Math.max( n - unevenmaxcount, 0 );
				
		return	excess;
	}
	
	//===============================================================================
	//	個体（染色体、、、）
	//===============================================================================
	class Individual implements Comparable<Individual> {
		Rectangle[]		sequence;					// 染色体（長方形の並び）
		char			operation;					// I,S,C,M　　生成起源
		int				point;						// 難易度ポイント（推定後）
		int				age;						// 世代
		int				orgpoint;					// Eval仮ポイント
		int				cmprange = -100;			// レンジ比較  -1:未満、0:範囲内、1:以上
	
		// 目的関数値関連　　変数名が　*G　は、重みを掛けたもの
		int				restcount,		restG;		// 未決定セル数
		int				mdivcount,		mdivG;		// 局所２重解の個数
		int				rectshapeeval, rectshapeG;	// 長方形形状制約
		int				prefixeval,	prefixG;		// 隠し形状
		int				prenumeval,	prenumG;		// 固定数字
		int				uneveneval,	unevenG;		// 偏りカウント
		
		double			time;

		Individual( ArrayList<Rectangle> reclist, char ope ) {
			if( reclist.size() != peglength ) {
				System.out.println("============= Gene, !!! Fatal error !!! length error" );
				return;
			}
			sequence = reclist.toArray(new Rectangle[reclist.size()]);
			operation = ope;
		}
		
		Individual( Rectangle[] recarray ) {
			sequence = recarray;
		}
		
		public		ArrayList<Rectangle>		getRectList() {
			return new ArrayList<Rectangle>(Arrays.asList(sequence));
		}
		
		private	int		eval() {
			return	evalbase();
		}
		
		private	double		evalwithage() {
			double	base = evalbase();
			double	agepart = (generation - this.age)*agingfactor;		// 老化！
			return	base + agepart;
		}
		
		private	int		evalbase() {
			return	restG + mdivG + rectshapeG + unevenG + prenumG + prefixG 
							+ Math.abs(cmprange);
		}

		public int compareTo(Individual that) {
			double		diff = this.evalwithage() - that.evalwithage();
			if( diff < 0 )
				return	-1;
			if( diff > 0 )
				return	+1;
			int	adiff = that.age - this.age;
			return	adiff;
		}
		
		public	void restore() {
			gridfield.clearRect();
			gridfield.clearAllArea();
			
			ArrayList<Rectangle> list = new ArrayList<Rectangle>();
			for( Rectangle rec : sequence ) {
				list.add(rec);
			}
			gridfield.setAreas( list );
		}
		
		public	void	pushRecsonGrid() {
			gridfield.genrecs = sequence;
		}
				
		public	String	getInovMark() {
			int	val = eval();
			if( val < minselect )
				return	"**";
			if( val < maxselect )
				return	"*";
			return	"";
		}
		
		public	void	updatecmprange() {
			int		val = 0;
			double	ptlog = Math.log(point);
			double	ptroot = Math.sqrt(point);
			if( lowerlimit > point ) {
				double	lowlog = Math.log(lowerlimit);
				double diff = ptlog - lowlog;
				val = -1 + (int)(diff*ptroot*limitweight);
			} else if( upperlimit <= point ) {
				double	uplog = Math.log(upperlimit);
				double diff = ptlog - uplog;
				val = 1 +(int)(diff*ptroot*limitweight);
			}
			
			cmprange = val;
		}
		
		private	void	reeval() {
			updatecmprange();		// cmprange 再評価
			restG      = restcount==0     ? 0 : Math.max( (int)(restcount*restcountweight), 1 );
			mdivG	    = mdivcount==0 || mdivcountweight==0.0 
									? 0 : Math.max( (int)(mdivcount*mdivcountweight), 1 );
			rectshapeG = rectshapeeval==0 ? 0 : Math.max( (int)(rectshapeeval*rectweight), 1 );	
			unevenG    = uneveneval==0    ? 0 : Math.max( (int)(uneveneval*unevenweight), 1 ) ;
			prenumG    = prenumeval==0    ? 0 : Math.max( (int)(prenumweight*prenumeval), 1 );
			prefixG    = prefixeval==0    ? 0 : Math.max( (int)(prefixweight*prefixeval), 1 );
			
			gridfield.difficultypanel.setunevencount(uneveneval);
		}
	
		public	String toString() {
			String strmdiv  = mdivcountweight!=0 ? String.format("%4d ", mdivG ) : "";
			String	strrect  = isrectprint  ? String.format("%4d ", rectshapeG  ) : "";
			String strsuneven = isunevenprint ? String.format("%4d ", unevenG ) : "";
			String strnum   = isnumprint   ? String.format("%4d ", prenumG ) : "";
			String strhid   = ishidprint   ? String.format("%6d ", prefixG ) : "";
			String strrng   = isrngprint   ? 
					String.format( "%4d : %6d %6d ", cmprange, point, orgpoint) :
					String.format( " : %6d ", restcount==0 ? point : orgpoint );
			String	strtm    = istimeprint  ? String.format(" : %6.3f", time ) : "";
					
			return	String.format("%4d%c %6d : %4d %s%s%s%s%s%s%s", age, operation, eval(), 
						restG, strmdiv, strrect, strsuneven, strnum, strhid, strrng, strtm );
					
		}
	}

	//===============================================================================
	//	分割処理
	//===============================================================================
	class Cutting {
		private	Rectangle		area;
		private	int				offsetx;
		private	int				offsety;
		private	int				width;
		private	int				height;
		private	Island[][]		belong;
		private	boolean[][]		blank;
		private	boolean[][]		near;
		private	ArrayList<Point>	nearlist;

		private	Island[]	islands;
	
		public Cutting ( int w, int h ) {
			offsetx = offsety = 0;
			width  = w;
			height = h;
			area = new Rectangle(offsetx,offsety,width,height);
		}
		
		public	Cutting ( int x, int y, int w, int h ) {
			offsetx = x;
			offsety = y;
			width  = w;
			height = h;
			area = new Rectangle(offsetx,offsety,width,height);
		}

		public ArrayList<Rectangle> getRects() {
			ArrayList<Rectangle> recs = new ArrayList<Rectangle>();
			for( Island is : islands ) {
				recs.add( (Rectangle)is.rec.clone() );
			}
			return	recs;
		}
		
		//		boolean[][] pat		パターン
		//		Rectangle[] recs		長方形確定済 交叉、突然変異で使用  変更のない長方形
		void	precut( boolean[][] pat, Rectangle[] recs, boolean ismutate ) {
			int prepegn = width * height;
			
			for(;;) {
				precutexec(pat);

				if( recs!=null ) {
					Rectangle[] currecs = new Rectangle[recs.length];
					for( int i=0; i<recs.length; ++i )
						currecs[i] = recs[i];
					
					// 突然変異のため、一部破壊　(0～4個破壊)
					if( ismutate ) {	
						for( int i=0; i<4; ++i ) {
							int	r = Util.rand.nextInt(recs.length);
							currecs[r] = null;
						}					
					}					
					for( Rectangle rec : currecs ) {
						if( rec == null )
							continue;
						prefit( rec, false );
					}
				}
				
				boolean	st = expandRecs(ismutate);
				
				if( st )
					break;

				int	pegn = countNoBelong();
				if( pegn == 0 || pegn==prepegn )
					break;
				prepegn = pegn;
			}
		}
		
		// 指定長方形の形を事前設定する
		void prefit( Rectangle rec, boolean pre ) {
			if( rec == null )
				return;
			
			Island	il = null;
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy ) {
					Island wk = belong[rec.x+dx][rec.y+dy];
					if( wk != null )
						il = wk;
				}
			if( il == null )		// error!
				return;

			if( pre ) {
				il.preset = true;
			}
			finishOneRect( il, rec );
		}
		
		void	precutexec( boolean[][] pat ) {
			pattern = pat;
			belong = new Island[width][height];

			int	pegn = pegcount();

			islands = new Island[pegn];
			initIslands();
		}

		private	boolean	expandRecs( boolean ismutate ) {		
			expandMust();
			
			if( ! expand2() )
				return false;
			
			expand2after();

			setupNear();
			
			int n = 0;
			for(int i=0;i<10;++i) {
				expandMust();
				expandNear();	

				if( stage==GAstage.Cross && i<1 ) {
					checkMultiDiv();
					expand2();
					expand2after();
				}
				
				setupBlank();
				breakNearBlank();			
				n = countNoBelong();
				if( n == 0 )
					break;
			}
						
			return	n==0;			
		}
		
		private	int		pegcount() {
			int	cnt = 0;
			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y )
					if( pattern[x][y] )
						++cnt;

			return	cnt;
		}

		// 島の初期生成（ペグだけを含む島を作る）
		private	void	initIslands() {
			int	cnt = 0;
			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y ) 
					if( pattern[x][y] ) {					
						belong[x][y] = islands[cnt] = new Island(x,y);
						++cnt;	
					}
		}

		// すべての島を、２のサイズに拡大する。
		private boolean	expand2() {
			for(;;) {
				for( Island is : islands ) {
					if( is.size > 1 )
						continue;
					int	initr = Util.rand.nextInt(4);
					boolean	found = false;
					for( int i=0; i<4; ++i ) {
						Point	dir;
						dir = getDirection( (initr+i) % 4 );
						int x = is.peg.x + dir.x;
						int y = is.peg.y + dir.y;
						if( ! ison(x,y) )
							continue;
						if( pattern[x][y] )
							continue;
						found = true;
						if( belong[x][y]!=null && belong[x][y].size==2 ) {
							Island bi = belong[x][y];
							bi.rec = new Rectangle( bi.peg.x, bi.peg.y, 1,1 );
							bi.size = 1;
							belong[x][y]=null;
						}
						belong[x][y] = is;
						is.rec = Util.makeRectangle( is.peg, new Point(x,y) );
						is.size = 2;
						break;
					}
					if( ! found )
						return	false;
					
					expandMust();									// １つ決まる毎に必然チェック
				}

				boolean	all2 = true;
				for( Island is : islands ) {
					if( is.size < 2 )
						all2 = false;
				}
				if( all2 )
					break;
			}
			
			return	true;
		}

		// ２のサイズに拡大の後処理
		private	void	expand2after() {
			for( int i=0; i<10; ++i ) {		//  10回も繰り返せば十分だろう
				ArrayList<Island>	list = new ArrayList<Island>();
				for( Island is : islands ) {
					if( is.size == 2 )
						list.add(is);
				}
				if( list.size() == 0 )
					break;
				while( list.size() > 0 ) {
					int	idx = Util.rand.nextInt(list.size());
					Island target = list.get(idx);
					list.remove(idx);
					Point  tp = target.peg;				// ペグの位置
					for( Island iq : list ) {
						if( iq.preset )
							continue;
						Point tq = iq.peg;
						if( Math.abs(tp.x - tq.x)!=1 || Math.abs(tp.y - tq.y)!=1 )
							continue;
						Rectangle unirec = target.rec.union(iq.rec);
						if( unirec.width==2 && unirec.height==2 ) {
							expand2bit(target);
						}
					}
				}
			}
		}
				
		// 面積２の島を少しだけ拡大する
		void	expand2bit( Island target) {
			int		r = Util.rand.nextInt(4);
			for( int i=0; i<4; ++i ) {			// ４方向試してみる
				Point	dir = getDirection( (r+i) % 4 );
				Rectangle	widenrec = new Rectangle(target.rec);
				if( dir.x==0 ) {		//  縦方向に伸ばす
					if( dir.y== -1 )
						-- widenrec.y;
					++widenrec.height;
				}
				if( dir.y==0 ) {		//  縦方向に伸ばす
					if( dir.x== -1 )
						-- widenrec.x;
					++widenrec.width;
				}
				if( widenrec.x<0 || widenrec.y<0 
						|| widenrec.x+widenrec.width>width || widenrec.y+widenrec.height > height)
					continue;
				boolean	ok = true;
				for( int dx=0; dx<widenrec.width; ++dx )
					for( int dy=0; dy<widenrec.height; ++dy ) {
						int	x = widenrec.x+dx;
						int	y = widenrec.y+dy;
						Island	bi = belong[x][y];
						if( bi!=null && bi!=target )
							ok = false;
					}
				if( !ok )
					continue;
				
				target.rec = widenrec;
				target.size = target.rec.width * target.rec.height;
				for( int dx=0; dx<widenrec.width; ++dx )
					for( int dy=0; dy<widenrec.height; ++dy )
						belong[widenrec.x+dx][widenrec.y+dy] = target;	
				return;
			}
		}

		void	setupNear() {
			near = new boolean[width][height];

			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y ) {
					if( belong[x][y] != null )
						continue;
					boolean	neighbor = false;
					for( int i=0; i<4; ++i ) {
						Point dir = getDirection(i);
						int	u = x + dir.x;
						int	v = y + dir.y;
						if( !ison(u,v) )
							continue;
						Island ild = belong[u][v];
						if( ild!=null && !ild.preset ) {
							neighbor = true;
							break;
						}
					}
					if( neighbor ) {
						near[x][y] = true;
					}
				}

			setupNearList();
		}

		private	void	setupNearList() {
			nearlist = new ArrayList<Point>();

			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y )
					if( near[x][y] )
						nearlist.add(new Point(x,y));
		}
		
		private	void	expandMust() {
			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y ) {
					Island is = belong[x][y];
					if( is != null )
						continue;
					// cell(x,y)は所属が決っていない。
					int	covercount = 0;
					Point	c = new Point(x,y);
					Island pgis = null;

					if( coverpegs[x+area.x][y+area.y] == null ) {
						System.out.printf(" (%d,%d) offset(%d,%d)%n", x,y,area.x,area.y);
						continue;
					}
					for( Point pg0 : coverpegs[x+area.x][y+area.y] ) {
						if( !area.contains(pg0) )
							continue;
						Point	pg = new Point(pg0.x-area.x,pg0.y-area.y);
						Rectangle	rec = Util.makeRectangle(c,pg);
						if( belong[pg.x][pg.y] == null )
							continue;
						if( belong[pg.x][pg.y].preset )
							continue;
						if( checkCover( rec, pg ) ) {
							++covercount;
							pgis = belong[pg.x][pg.y];
						}
					}
					if( covercount == 1 ) {
						expandOneRect( c, pgis );
					}
				}
		}
		
		private	boolean	checkCover( Rectangle rec, Point pg ) {
			Island	pgis = belong[pg.x][pg.y];

			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy ) {
					int x = rec.x+dx;
					int y = rec.y+dy;
					Island be = belong[x][y];
					if( be!=null && be!=pgis  ) {
						return	false;
					}
				}
			return	true;
		}
		
		// 複数分割がある場合を調べる。
		private void checkMultiDiv() {
			ArrayList<Rectangle> destareas = new ArrayList<Rectangle>();
			
			check2x4MultiDiv( destareas );
			check342MultiDiv( destareas );
			check333MultiDiv( destareas );
			
			ArrayList<Island> destislands = new ArrayList<Island>();

			for( Rectangle area : destareas ) {
				for( Island is : islands ) {
					Rectangle sec = area.intersection(is.rec);
					if( ! (sec.width>0 && sec.height>0 ) )
						continue;
					if( destislands.indexOf(is) >= 0 )
						continue;
					destislands.add(is);
				}
			}

			while( destislands.size() > 0 ) {
				int	idx = Util.rand.nextInt(destislands.size());
				Island is = destislands.get(idx);
				destislands.remove(idx);
				
				is.minimize();				// 破壊、最小サイズ１にする
			}
		}
		
		private void	check2x4MultiDiv( ArrayList<Rectangle> destareas ) {
			if( destareas.size() > 1 )
				return;
			
			for( Island is1 : islands ) {
				if( is1.size != 4 )
					continue;
				Point	p1 = is1.peg;
				for( Island is2 : islands ) {
					if( is2.size != 4 )
						continue;
					Point p2 = is2.peg;
					if( ! (p2.x > p1.x  && p2.y != p1.y) )
						continue;
					Rectangle runi = is1.rec.union(is2.rec);
					if( runi.width * runi.height != 8 )
						continue;
					Rectangle half = new Rectangle(runi);
					if( half.width == 4 )
						half.width = 2;
					else
						half.height = 2;
					if( (half.contains(p1) ? 1 : 0) + (half.contains(p2) ? 1 : 0) != 1 )
						continue;
					
					if( runi.width==4 ) {
						runi.y -= 1;	runi.height = 4;
					} else {
						runi.x -= 1;	runi.width = 4;
					}
					destareas.add( runi );

					return;		
				}
			}
		}

		//  3-4-2 diagonal  3x3
		private void	check342MultiDiv( ArrayList<Rectangle> destareas ) {
			if( destareas.size() > 1 )
				return;
			
			for( Island is1 : islands ) {
				if( is1.size != 4 )
					continue;
				Point	p1 = is1.peg;
				for( Island is2 : islands ) {
					if( is2.size != 3 )
						continue;
					Point p2 = is2.peg;
					if( ! ( Math.abs(p1.x-p2.x)==1 && Math.abs(p1.y-p2.y)==1 ) )
						continue;
					Rectangle runi = is1.rec.union(is2.rec);
					if( ! ( runi.width==3 && runi.height==3 ) )
						continue;
					for( Island is3 : islands ) {
						if( is3.size != 2 )
							continue;
						Rectangle r3 = runi.union(is3.rec);
						if( ! runi.equals(r3) )
							continue;
						
						if( Util.rand.nextDouble() < 0.5 ) {
							runi.y -= 1;	runi.height = 5;
						} else {
							runi.x -= 1;	runi.width = 5;
						}
						destareas.add( runi );	

						return;
					}
				}
			}
		}
		
		//  3-3-3 diagonal  3x3
		private void	check333MultiDiv( ArrayList<Rectangle> destareas ) {
			if( destareas.size() > 1 )
				return;
			
			for( Island is1 : islands ) {
				if( is1.size != 3 )
					continue;
				Point	p1 = is1.peg;
				for( Island is2 : islands ) {
					if( is2.size != 3 )
						continue;
					Point p2 = is2.peg;
					if( p2.x != p1.x+1 || p2.y==p1.y )
						continue;
					Rectangle runi = is1.rec.union(is2.rec);
					if( runi.width * runi.height > 9 )
						continue;
					for( Island is3 : islands ) {
						if( is3.size != 3 )
							continue;
						Point p3 = is3.peg;
						if( p3.x != p1.x+2 || p3.y==p1.y || p3.y==p2.y )
							continue;
						runi = runi.union(is3.rec);
						if( !(runi.width==3 && runi.height==3 ) )
							continue;
						
						if( Util.rand.nextDouble() < 0.5 ) {
							runi.y -= 1;	runi.height = 5;
						} else {
							runi.x -= 1;	runi.width = 5;
						}
						destareas.add( runi );		
						
						return;
					}
				}
			}
		}
		
		private void expandNear() {
			int	prev = countNoBelong();

			for( int lpn = 0; lpn < 2; ++lpn ) {
				expandNearAll();
				int	current = countNoBelong();
				if( current == 0 )
					break;
				if( current < prev ) {
					prev = current;
					lpn = 0;
				}
				setupNear();
			}
		}

		private	void	expandNearAll() {
			ArrayList<Point>	list = new ArrayList<Point>();
			while( nearlist.size() > 0 ) {
				int	idx = Util.rand.nextInt(nearlist.size());
				list.add(nearlist.get(idx));
				nearlist.remove(idx);
			}
			nearlist = list;

			for( Point p : nearlist ) {
				expandNearOne(p);
			}
		}

		//	近所を１箇所だけ処理する
		private	void	expandNearOne( Point p ) {	
			ArrayList<Island>	isls = new ArrayList<Island>();
		
			for( int i=0; i<4; ++i ) {
				Point	dir = getDirection(i);
				int	u = p.x + dir.x;
				int	v = p.y + dir.y;
				
				for(;;) {
					if( !ison(u,v) )
						break;
					Island isl = belong[u][v];
					if( isl != null && !isl.preset) {
						isls.add( isl );
						break;
					}
					u += dir.x;
					v += dir.y;
					if( stage == GAstage.Pre ) {	// 前処理の時は、隣だけ処理する。
						if( cuttrycount == 0 )
							continue;
						if( Util.rand.nextFloat() > 5.0/(6.0+cuttrycount) )		// 確率自動調整
							break;
					}
				}
			}
			int	n = isls.size();
			if( n == 0 )	// 失敗
				return;

			int		rnd = Util.rand.nextInt(n);

			for( int i=0; i<n; ++i ) {
				Island	is = isls.get((rnd+i)%n);				// 拡張する島
				
				if( expandOneRect( p, is ) ) {
					expandMust();									// １つ決まる毎に必然チェック
					return;
				}
			}		
		}
		
		//	セルp, 島　を含む拡張長方形に拡張する。
		//	成功したとき、true, そうでないとき false(何もしない)
		private boolean	expandOneRect( Point p, Island is ) {
			Rectangle	rec = Util.makeRectangle(is.rec, p);	// 拡張された島
			
			ArrayList<Island>  cross = new ArrayList<Island>();	// 縮小する島
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy) {
					Island cis = belong[rec.x+dx][rec.y+dy];
					if( cis == null || cis == is )
						continue;
					if( cross.indexOf(cis) >= 0 )
						continue;
					cross.add(cis);
				}
			
			if( cross.size()!=0 )			// 他の島と被る
				return	false;

			finishOneRect( is, rec );
			
			return	true;
		}

		private	void	finishOneRect( Island is, Rectangle rec ) {
			is.rec = rec;
			is.size = rec.width*rec.height;
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy) {
					belong[rec.x+dx][rec.y+dy] = is;
					if( near != null )
						near[rec.x+dx][rec.y+dy] = false;
				}
		}
		
		//	空白地帯をblank[][]に設定する
		private	void	setupBlank() {
			blank = new boolean[width][height];
			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y )
					if( belong[x][y] == null )
						blank[x][y] = true;
		}

		private	void	breakNearBlank() {
			ArrayList<Island>  list = new ArrayList<Island>();

			for( int x=0; x<width; ++x )
				for( int y=0; y<height; ++y ) {
					if( ! blank[x][y] )
						continue;
					for( int i=0; i<4; ++i ) {
						Point dir = getDirection(i);
						int u = x + dir.x;
						int v = y + dir.y;
						if( !ison(u,v) )
							continue;
						Island is = belong[u][v];
						if( is == null || is.preset )
							continue;
						if( list.indexOf(is) >= 0 )
							continue;
						list.add(is);		
					}
				}

			for( Island is : list ) {
				is.minimize();				// 破壊、最小サイズ１にする
			}
		}

		private	int		countNoBelong() {
			int	cnt = 0;
			for( int x=0; x<width; ++x ) {
				for( int y=0; y<height; ++y ) {
					if( belong[x][y] == null )
						++cnt;
				}
			}
			return	cnt;
		}
		
		boolean	ison( int x, int y ) {
			return	x>=0 && x<width && y>=0 && y<height;
		}

		Point	getDirection( int i ) {
			final	int[]	xdir = { 1, 0,-1, 0 };
			final	int[]	ydir = { 0, 1, 0, -1};

			return	new Point(xdir[i],ydir[i]);
		}
	
		class	Island { 
			Rectangle	rec;		// 長方形
			Point		peg;		// ペグ（数字）の位置
			int			size;		// 面積
			boolean	preset;	// 初期指定長方形

			Island( int x, int y ) {
				peg = new Point(x,y);
				rec = new Rectangle(x,y,1,1);
				size = 1;
			}

			void minimize() {
				for( int dx=0; dx<rec.width; ++dx)
					for( int dy=0; dy<rec.height; ++dy ) {
						int u = rec.x+dx;
						int v = rec.y+dy;
						belong[u][v] = null;
					}
				belong[peg.x][peg.y] = this;
				rec = new Rectangle(peg.x,peg.y,1,1);
				size = 1;				
			}
		}
	}
	
	private	void	clearlog() {
		gridfield.makepanel.clearlog();
	}
	
	private	void	appendlog( String str ) {
		gridfield.makepanel.appendlog(str);	
	}
	
	private	void	updateexectime() {
		long currenttime = System.currentTimeMillis();
		
		int	sec = (int)(currenttime - starttime) / 1000;
		gridfield.makepanel.setexectime(sec);
	}

	//===============================================================================
	//	染色体プールの履歴情報
	//===============================================================================

	public		ArrayList<GeneticPoolInfo>	geneinfo;
	
	public	class	GeneticPoolInfo {
		int		selectmin;
		int		selectmax;
		double	average;
		int		mutatecount;
		
		GeneticPoolInfo()	{}
		GeneticPoolInfo( int min, int max, double av, int mc ) {
			selectmin	= min;
			selectmax	= max;
			average	= av;
			mutatecount = mc;
		}
	}
	
	private	void	addGeneInfo( int min, int max, double av ) {
		GeneticPoolInfo gi = new GeneticPoolInfo( min, max, av, realmutatecount );
		geneinfo.add(gi);
		gridfield.graphpanel.update();
	}

	//==========================================================================
	//		評価ヒストグラム
	//==========================================================================
	/*
	class	EvalHist	{
		private	ArrayList<EvalCount>	list;
		
		EvalHist() {
			list = new ArrayList<EvalCount>();
		}
			
		void	add( int e ) {
			EvalCount	found=null;
			for( EvalCount ec : list )
				if( ec.eval == e ) {
					++ec.count;
					found = ec;
					break;
				}	
			if( found == null  )
				list.add( found = new EvalCount(e) );

			return;	
		}
		
		int		getcount( int e ) {
			for( EvalCount ec : list )
				if( ec.eval == e )
					return	ec.count;
			return	0;
		}
		
		class	EvalCount	{
			int		eval;
			int		count;

			EvalCount( int e )	{	eval = e;	count = 1;	}
		}
	}
	*/
}
