package numarea;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Vector;

import javax.swing.JLabel;

import numarea.DesignPanel;

public class GridField extends JLabel implements MouseListener , MouseMotionListener, MouseWheelListener , KeyListener
{
	private static final long serialVersionUID = 1L;

	MainPanel		mainpanel;
	DesignPanel	designpanel = null;
	DifficultyPanel difficultypanel = null;
	MakePanel		makepanel = null;
	GraphPanel		graphpanel = null;
	ShapePanel		shapepanel = null;
	
	private	boolean	lock;			// 表示用ロック
	
	private Rectangle rect;				// 長方形カーソル
	private Point rectStart;
	private Point rectEnd;

	private int bwidth, bheight;		// 盤面サイズ
	private int unit = 32;				// マスのサイズ
	private int x0 = 15 , y0 = 15;		// 表示位置オフセット
	
	public	int	getBWidth()		{	return	bwidth;		}
	public	int	getBHeight()		{	return	bheight;		}
	
	private boolean[][] patterntbl;		// ○の場所
	private int[][] areatbl;			// 面積のテーブル（問題）
	private boolean[][]	preshapetbl;	// プレ形状
	private int[][]	preshapeweight;
	
	private int[][] prenumtbl;			// プレ数値
	
	private boolean[][] copybuffer = null;		// コピー用のバッファ
	private int[][] copyarea = null;			// 面積コピー
	
	private ArrayList<Rectangle> arealist;		// 長方形リスト
	private ArrayList<Rectangle> arealistsaved;
	
	private int[][] playtbl;				// 遊ぶときに入れる値を保持
	private int[][] idtbl;				// Solverで解いたときのidテーブル
	private ArrayList<Rectangle> isrecs;		// Solverで解いたときの矩形テーブル
	private Image buffer ;
	private Font numberFont ;
	public	ArrayList<Rectangle> prerecs;	// 初期形状
	
	public		GA		ga = new GA(this);
	public		Solver solver;
		
	public		boolean		generating;	// 自動生成中
	public		Rectangle[]	genrecs;
	
	private	boolean	isbreak = false;
	public 	void		setBreak( boolean b )	{	isbreak = b;		}
	public		boolean	isBreak() 					{	return	isbreak;	}
	
	public		boolean	useInitfixhalf	= true;
	public		boolean	usePegfixhalf		= true;
	public		boolean	useCelluniqpeg	= true;
	public		boolean	useCellcommon		= true;
	public		boolean	useRegiondivide	= false;
	
	public		boolean	historyon			= false;
	
	private int	difficultypoint = 0;
	private double	difficultyav = 0.0;
	
	SymmetricProperty symmetry = new SymmetricProperty();
	private boolean[][] symError; // Symmetric Violation
	
	private boolean	skippaint = false;
	public	void	setskippaint( boolean b )	{	skippaint = b;	}
	
	private Point cursor ;
	
	private static Color ColorLevel1		= new Color(16,96,0);
	private static Color ColorLevel2		= new Color(16,192,0);	
	private static Color ColorLevel3		= new Color(40,212,32);	
	private static Color ColorLevel4		= new Color(132,248,48);
	private static Color ColorLevel5		= new Color(152,216,48);	
	private static Color ColorLevel6		= new Color(216,216,48);	
	private static Color ColorLevel7		= new Color(244,160,48);		
	private static Color ColorLevel99		= new Color(255,0,0);		
	private static Color insideColorFixed	= Color.GREEN;
	private static Color borderColorFixed	= Color.BLACK;
	private static Color hiddenColor		= Color.MAGENTA;
	private static Color insideColorUndef	= new Color(246,192,64);
	private static Color borderColorUndef	= new Color(160,128,64);	
	private static Color borderColorUndefStrong	= new Color(200,128,32);
	private static Color normalPegColor	= Color.BLUE;
	private static Color hiddenPegColor	= Color.RED;
	
	public	Color	getNormalPegColor()	{	return	normalPegColor;	}
	public	Color	getFixedRecColor()	{	return	insideColorFixed;	}
	public	Color	getUndefRecColor()	{	return insideColorUndef;	}
	public	Color	getHiddenPegColor()	{	return	hiddenPegColor;	}
	public	Color	getHiddenRecColor()	{	return	hiddenColor;		}
	public	void	setNormalPegColor( Color c )	{	normalPegColor   = c;	}
	public	void	setFixedRecColor( Color c )		{	insideColorFixed = c;	}
	public	void	setUndefRecColor( Color c )		{	insideColorUndef = c;	}
	public	void	setHiddenPegColor( Color c )	{	hiddenPegColor   = c;	}
	public	void	setHiddenRecColor( Color c )	{	hiddenColor		= c;	}


	public GridField(MainPanel mainpanel) {
		this.mainpanel = mainpanel ; 
		setGridSize(10,10);	
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
		addKeyListener(this);
		this.requestFocus();
		
		mode = MODE.SETTING;
	}
	
	public MainPanel getMainPanel(){
		return this.mainpanel; 
	}

	public Dimension getGridSize(){
		return new Dimension(bwidth , bheight);
	}
	public	Solver	getSolver()	{	return	solver;	}
	
	public	int[][]	getDifficultytable() {
		return	solver.difficultylevel;
	}
	
	public	void	clearSolverButton() {
		mainpanel.clearSolveButton();
	}
	
	public	int		getDifficultypoint() {
		return	difficultypoint;
	}
	
	public	double	getDifficultyav() {
		return difficultyav;
	}
	
	public	void	setDifficultypoint(int pnt) {
		difficultypoint = pnt;	
		difficultyav = (double)pnt / (bwidth * bheight);

		mainpanel.setPoint( difficultypoint, difficultyav );
	}
		
	public	int[][]	getQuestionTable() {
		return	areatbl;
	}
	
	public	boolean[][]	getPatternTable() {
		return	patterntbl;
	}
	
	public	int[][]	getPreNumTable() {
		return	prenumtbl;
	}
	
	public	void	setPreNumTable( int[][] tbl ) {
		prenumtbl = tbl;
	}

	public	boolean[][]	getPreShapeTable() {
		return	preshapetbl;
	}
	
	public	int[][]	getPreShapeWeight() {
		setupPreShape();
		return	preshapeweight;
	}
	
	public	ArrayList<Rectangle>	getAnswerRectangles() {
		return	solver.answerrecs;
	}
	
	public	void	setAreas( ArrayList<Rectangle> recs ) {
		clearRect();
		clearAllArea();
		
		clearAreaList();
		for( Rectangle rec : recs ) {
			addOneArea( rec );
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy ) {
					int u = rec.x + dx;
					int v = rec.y + dy;
					if( patterntbl[u][v] ) {
						areatbl[u][v] = rec.width * rec.height;
					}
				}
		}
	}
	
	public	int	getBlankCount() {
		boolean[][]	tbl = new boolean[bwidth][bheight];
		
		// ペグの位置は つまっている　ことにする。
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y )
				tbl[x][y] = patterntbl[x][y];
		
		for( Rectangle rec : arealist) {
			if( rec == null )
				continue;
			if( howManyHint(rec) != 1 )
				continue;
			for( int dx=0; dx<rec.width; ++dx )
				for( int dy=0; dy<rec.height; ++dy )
					tbl[rec.x+dx][rec.y+dy] = true;
		}
		
		int	cnt = 0;
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y )
				if( ! tbl[x][y] )
					++cnt;
		
		return	cnt;
	}
	
	public ArrayList<Rectangle> getFixRectangles() {
		ArrayList<Rectangle> list = new ArrayList<Rectangle>();
		
		for( Rectangle rec : arealist) {
			if( rec == null )
				continue;
			if( howManyHint(rec) != 1 )
				continue;	
			list.add(rec);
		}
		return	list;
	}
	
	public ArrayList<Rectangle> getShapeOkRectangles() {
		ArrayList<Rectangle> list = new ArrayList<Rectangle>();
		
		for( Rectangle rec : arealist) {
			Point	peg = getPeg( rec );
			if( peg == null )
				continue;
			if( difficultypanel.isOkShape( rec, !Util.isInside( peg, bwidth, bheight )))
				list.add(rec);
		}
		
		return	list;
	}
	
	private	void	setupPreShape() {	
		preshapeweight = new int[bwidth][bheight];
		
		Point[] dirs = { new Point(-1,0), new Point(1,0), new Point(0,-1),new Point(0,1)};
		
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y ) {
				boolean p = preshapetbl[x][y];
				for( Point dir : dirs ) {
					int u = x+dir.x;
					int v = y+dir.y;
					if( u<0 || u>=bwidth || v<0 || v>=bheight )
						continue;
					if( p != preshapetbl[u][v])
						++preshapeweight[x][y];
				}
			}
		
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y ) {
				int v = preshapeweight[x][y];
				preshapeweight[x][y] = v > 0 ? v : 1;
			}
	}
	
	public	void	clearPreShape() {
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y )
				preshapetbl[x][y] = false;		
	}
	
	public	void	clearPreNumber() {
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y )
				prenumtbl[x][y] = 0;			
	}
	
	public	Point	getPegPos( Rectangle rec ) {
		for( int dx=0; dx<rec.width; ++dx)
			for( int dy=0; dy<rec.height; ++dy ) {
				if( patterntbl[rec.x+dx][rec.y+dy] )
					return	new Point(rec.x+dx,rec.y+dy);
			}
		return	null;
	}
	
	// -------------------------------------------------------------------------	
	//                          MODE
	// -------------------------------------------------------------------------
	public static enum MODE{
		INIT , 
		SETTING    ,
		CUTTING    ,
		HIDDEN		,
		NUMBER		,
		SOLVE     ,
		EVAL,
		ANSWER , 
		PLAY , 
	};
	public String toString(MODE m) {
		String str ="other";
		if( m == MODE.SETTING )	str = "SETTING";
		if( m == MODE.CUTTING )	str = "CUTTING";
		if( m == MODE.HIDDEN )	str = "HIDDEN";
		if( m == MODE.NUMBER )	str = "NUMBER";
		if( m == MODE.SOLVE )	str = "SOLVE";	
		if( m == MODE.EVAL )		str = "EVAL";
		if( m == MODE.PLAY )		str = "PLAY";		
		return	str;
	}
	private MODE mode = MODE.INIT;
	
	public void setMode(MODE m) {
		setMode( m, false );
	}
	public	MODE getMode()	{	return	this.mode;		}
	
	public void setMode(MODE m, boolean pass){
		MODE	prevmode = this.mode;
		this.mode = m;
		
		if( this.mode == MODE.SETTING ) {
			getMainPanel().setButton.setEnabled(true);
			getMainPanel().cutButton.setEnabled(true);				
			getMainPanel().playButton.setEnabled(false);
			getMainPanel().modeLabel.setText("SET");
			if( ! pass ) {
				isrecs = new ArrayList<Rectangle>();
			}
			
		}else if(this.mode == MODE.CUTTING) {
			getMainPanel().setButton.setEnabled(true);
			getMainPanel().cutButton.setEnabled(true);				
			getMainPanel().playButton.setEnabled(true);
			if( prevmode == MODE.SOLVE || prevmode == MODE.EVAL ) {
				setcut();
			}
			rect = null;
			getMainPanel().modeLabel.setText("CUT");

		}else if(this.mode == MODE.NUMBER) {
			mainpanel.setButton.setEnabled(true);
			mainpanel.cutButton.setEnabled(true);				
			mainpanel.playButton.setEnabled(true);
			rect = null;
			getMainPanel().modeLabel.setText("NUM");
			
		}else if(this.mode == MODE.HIDDEN) {
			mainpanel.setButton.setEnabled(true);
			mainpanel.cutButton.setEnabled(true);				
			mainpanel.playButton.setEnabled(true);
	
			getMainPanel().modeLabel.setText("HID");
			
		}else if( this.mode == MODE.PLAY) {
			getMainPanel().setButton.setEnabled(true);
			getMainPanel().cutButton.setEnabled(true);				
			getMainPanel().playButton.setEnabled(true);	
			getMainPanel().modeLabel.setText("PLAY");
			if( prevmode == MODE.CUTTING ) {
				// 
			} else if( prevmode != MODE.SOLVE ) {
				isrecs = new ArrayList<Rectangle>();
				arealist = new ArrayList<Rectangle>();	
			}

		}else if( this.mode == MODE.SOLVE) {
			mainpanel.setButton.setEnabled(true);
			mainpanel.cutButton.setEnabled(true);				
			mainpanel.playButton.setEnabled(true);	
			mainpanel.solveButton.setBackground(null);
			mainpanel.evalButton.setBackground(null);
			getMainPanel().modeLabel.setText("SOLVE");
			if( ! pass ) {
				rect = null;
				clearBoard();
				solve(true);
			}
			
		}else if( this.mode == MODE.EVAL) {
			mainpanel.setButton.setEnabled(true);
			mainpanel.cutButton.setEnabled(true);				
			mainpanel.playButton.setEnabled(true);	
			mainpanel.solveButton.setBackground(null);
			mainpanel.evalButton.setBackground(null);
			mainpanel.modeLabel.setText("EVAL");
			if( ! pass ) {
				rect = null;
				clearBoard();
				solve(false);
			}
		}
	
		refresh();
	}

	//---------------------------------------------------------------------
	//	サイズ変更
	//---------------------------------------------------------------------
	boolean setGridSize(int w , int h)
	{
		if(w * h > 1000000)
			return false;
		
		this.bwidth  = w;
		this.bheight = h; 
		cursor = null;
		rect = null;

		patterntbl = new boolean[w][h];
		areatbl = new int[w][h];
		arealist = new ArrayList<Rectangle>();
		playtbl = new int[w][h];
		symError = new boolean[w][h];
		preshapetbl = new boolean[w][h];
		prenumtbl = new int[w][h];
				
		solver = new Solver(w,h);
		
		setMode(MODE.SETTING);

		adjustSize();
		
		if( designpanel != null ) {
			designpanel.setInputSize(bwidth,bheight);
		}
		
		System.gc();
		repaint();
		
		return	true;
	}
	
	
	public void paint(Graphics g) {		
		if(bwidth==0 || bheight==0) return ;
		if(buffer == null){
			try{
				buffer = createImage(getWidth() , getHeight());
			}catch(OutOfMemoryError e){
				buffer = null;
				System.gc();
				zoomOut();
				return ;
			}
		}
		
		clearBuffer();
		drawSolved();
		drawAreaList();
		drawHidden();
		drawHints();
		drawSymErr();
		drawGrid();
		drawRectLine();
		
		Rectangle rect = getVisibleRect();
		
		g.drawImage(buffer , rect.x , rect.y , rect.x + rect.width , rect.y + rect.height 
				 , rect.x , rect.y , rect.x + rect.width , rect.y + rect.height , this);
		

		mainpanel.setPegNum(	countPegs() );
		
		lock = false;
	}
	
	//----------------------------------------------------------------------------
	// refresh
	//----------------------------------------------------------------------------	
	public	void	prepareRefresh() {
		for( int lc=0; lock && lc < 50; ++lc ) {
			try {	Thread.sleep(20);	} catch (InterruptedException e) {}
		}
		lock = true;
	}
		
	public void refresh(){
		if( skippaint )
			return;
		
		prepareRefresh();
		repaint();
	}
	
	public	void	forcedrefresh() {
		prepareRefresh();
		repaint();		
	}

	//============================================================================
	// Cell の表示
	//============================================================================
	public void update(Graphics g) {
		paint(g);
	}
	public void zoomIn(){
		double area = getPreferredSize().width * getPreferredSize().height * (9.0/4.0);
		if(area > 18000000.) return ;
		unit = unit * 5 / 4;
		adjustSize();
		repaint();
	}
	public void zoomOut(){
		unit = unit * 4 / 5 ;
		if(unit<=2) unit = 2;
		adjustSize();
		repaint();
	}
	public void zoom(int delta) {
		unit += delta;
		adjustSize();
		repaint();
	}
	private void cursorMove( int dx, int dy ) {
		int x = (cursor.x + bwidth - 1 + dx) % (bwidth-1);
		int y = (cursor.y + bheight - 1 + dy) %(bheight-1);
		cursorMoveTo( x, y );
	}
	public void cursorMoveTo(int x , int y){
		cursor = new Point(x,y);
		repaint();
	}
	
	void adjustSize(){
		this.setPreferredSize(new Dimension(2*x0 + unit * bwidth-1, 2*y0 + unit * bheight ));
		numberFont = new Font("SansSerif", Font.PLAIN, unit*2/3);	
		revalidate();
		buffer=null;
	}
	private int toXIndex(int x){
		if(x < x0) return -1;
		int xi = (x-x0)/unit;
		return xi;
	}
	private int toYIndex(int y){
		if(y < y0) return -1;
		int yi = (y-y0)/unit;
		return yi;
	}
	
	//---------------------------------------------------------------------------
	//		矩形領域処理
	//---------------------------------------------------------------------------
	public	Point	getPeg( Rectangle rec ) {
		for( int x=0; x<rec.width; ++x )
			for( int y=0; y<rec.height; ++y )
				if( patterntbl[rec.x+x][rec.y+y] ) {
					return	new Point(rec.x+x, rec.y+y);
				}
		return	null;
	}
	
	public	void clearRect() {
		rect = null;
	}
	
	public void drawRectLine() {
		if( rect == null )
			return;
		if( mode == MODE.SETTING || mode == MODE.CUTTING || mode == MODE.NUMBER ) {	
			// 矩形領域内に○が一ヶ所だけだと、数字を設定する
			setHintNumber();
		}

		Graphics g = buffer.getGraphics();
		int x = x0 + unit * rect.x;
		int y = y0 + unit * rect.y;	
		g.setColor(Color.ORANGE);
		
		int w = 2;
		g.fillRect(x+1 , y+1 , w, unit*rect.height-1);	
		g.fillRect(x+1 , y+1 , unit*rect.width-1, w);		
		g.fillRect(x+unit*rect.width-w,y + 1, w, unit*rect.height-1);
		g.fillRect(x+1,y+unit*rect.height-w, unit*rect.width-1,w);
	}

	public void rectStartTo(int x, int y) {
		rect = new Rectangle(x,y,0,0);
		rectStart = new Point(x,y);
	}

	public void rectDraggedTo(int x, int y) {
		if( rect == null )
			return;
		rectEnd = new Point(x,y);
		
		int rx = Math.min(rectStart.x,rectEnd.x);
		int ry = Math.min(rectStart.y,rectEnd.y);
		int rw = Math.abs(rectStart.x - rectEnd.x) + 1;
		int rh = Math.abs(rectStart.y - rectEnd.y) + 1;
		
		Rectangle r = new Rectangle(rx,ry,rw,rh);

		if( r.equals(rect) )
			return;
		
		if( mode == MODE.NUMBER ) {
			if( howManyHint(r) == 1 ) {
				Point	pos = whereisHint(r);
				int	s = r.width * r.height;
				if( s <= 1 )
					s = 0;
				prenumtbl[pos.x][pos.y] = s;
			}
		} else {
			arealist = new ArrayList<Rectangle>(arealistsaved);
			addOneArea( r );
		}
		
		rect = r;
		drawRectLine();
		repaint();
	}
	
	// Solverで解いた結果を表示
	// 所属IDが決まった場合に、マスをGreeenにする。
	private void drawSolved() {
		if( ! (mode == MODE.SOLVE || mode == MODE.EVAL ) )
			return;
		
		Graphics g = buffer.getGraphics();
		
		// セルの色を塗る
		for( int i=0; i<bwidth; ++i )
			for( int j=0; j<bheight; ++j ) {
				int x = x0 + unit * i;
				int y = y0 + unit * j;
				int id = idtbl[i][j];
				if( id == 0 )
					continue;
				Color c = id > 0 ? insideColorFixed : insideColorUndef;
				g.setColor( c );
				g.fillRect( x+1, y+1, unit-1, unit-1 );
			}

		// 枠の色を塗る
		if( isrecs != null ) {
			for( Rectangle r : isrecs ) {
				if( r == null )
					continue;
				int x = x0 + unit * r.x;
				int y = y0 + unit * r.y;	

				g.setColor(borderColorFixed);		
				int w = 2;
				g.fillRect(x+1 , y+1 , w, unit*r.height-1);	
				g.fillRect(x+1 , y+1 , unit*r.width-1, w);		
				g.fillRect(x+unit*r.width-w,y + 1, w, unit*r.height-1);
				g.fillRect(x+1,y+unit*r.height-w, unit*r.width-1,w);	
			}
		}
	}
	
	private void drawAreaList() {
		if( this.mode == MODE.HIDDEN || this.mode == MODE.NUMBER )
			return;
		
		try {
			//		solver.printDifficultylevel();
			boolean usesolved = isrecs!=null && isrecs.size()!=0;
			if( usesolved ) {
				for( Rectangle r : isrecs ) {
					drawArea( r, true );
				}
			}
			if( arealist != null ) {
				for(Rectangle r : arealist ) {
					drawArea( r, !usesolved );
				}
			}

		} catch( Exception e ) {}
	}
	
	private void drawHidden() {
		if( this.mode != MODE.HIDDEN )
			return;
		
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y )
				if( preshapetbl[x][y] ) {	// HIDDEN
					drawArea( new Rectangle(x,y,1,1), true, false );
				}		
	}
	
	public	int	getShapeCount( ArrayList<Rectangle> list, int lside, int sside ) {
		int	cnt = 0;
		for( Rectangle rec : list ) {
			int	ls = Math.max( rec.width, rec.height );
			int	ss = Math.min( rec.width, rec.height );
			if( !( ls == lside && ss == sside ) )
				continue;

			++cnt;
		}
		
		return	cnt;
	}
	
	public	int	getShapeCount( ArrayList<Rectangle> list, int area, String areacmp ) {
		int	cnt = 0;
		for( Rectangle rec : list ) {
			int	ar = rec.width * rec.height;			
			if( areacmp.equals("=") ) {
				if( ar != area )
					continue;
			} else if( areacmp.equals( "<=") ) {	// ( area<=　)
				if( ar > area )
					continue;
			} else if( areacmp.equals( ">=") ) {	// ( area>=　)
				if( ar < area )
					continue;
			}
			++cnt;
		}
		
		return	cnt;
	}
	
	private	ArrayList<Rectangle>	getCurrentRectList() {
		if( generating && genrecs != null ) {	// 自動生成中
			ArrayList<Rectangle> list = new ArrayList<Rectangle>();
			for( Rectangle rec : genrecs )
				list.add(rec);
			return	list;
		}
		
		switch( mode ) {
		case SOLVE:
		case EVAL:
			return	getCheckedArealist( solver.getSequenceRectangles());
			
		case CUTTING:
			return	getCheckedArealist();
		}
		return	null;
	}
	
	private void drawArea( Rectangle r, boolean isfill ) {	
		drawArea( r, isfill, true );
	}
	
	private void drawArea( Rectangle r, boolean isfill, boolean isframe ) {	
		if( r == null )
			return;
		
		int n = howManyHint(r);
		Color insideColor = insideColorFixed;
		Color borderColor = borderColorFixed;
		if( mode == MODE.CUTTING ) {
			if( n != 1 )
				return;
		} else if( mode == MODE.PLAY ) {
			if( ! checkHint(r) ) {
				insideColor = insideColorUndef;
				borderColor = borderColorUndef;
			}
			Point p = whereisHint(r);
			if( n==1 && p!=null && preshapetbl[p.x][p.y] && checkHint(r) ) {
				insideColor = hiddenColor;
			}
			
		} else if( mode == MODE.SOLVE || mode == MODE.EVAL ) {
			if( ! checkHint(r) ) {	
				borderColor = borderColorUndefStrong;				
			}
		} else if( mode == MODE.HIDDEN ) {
			insideColor = hiddenColor;
		}
		
		Graphics g = buffer.getGraphics();
		int x = x0 + unit * r.x;
		int y = y0 + unit * r.y;
		
		if( isfill ) {
			g.setColor(insideColor);
			g.fillRect(x+1, y+1, unit*r.width-1, unit*r.height-1);
			drawAreaLevel( r );
		}
		
		if( isframe ) {
			g.setColor(borderColor);		
			int w = 2;
			g.fillRect(x+1 , y+1 , w, unit*r.height-1);	
			g.fillRect(x+1 , y+1 , unit*r.width-1, w);		
			g.fillRect(x+unit*r.width-w,y + 1, w, unit*r.height-1);
			g.fillRect(x+1,y+unit*r.height-w, unit*r.width-1,w);	
		}
	}
	
	private void setHiddenArea() {
		if( preshapetbl == null )
			return;
		
		for( int x=rect.x; x<rect.x+rect.width; ++x )
			for( int y=rect.y; y<rect.y+rect.height; ++y )
				preshapetbl[x][y] = !preshapetbl[x][y];
	}
	
	private void drawAreaLevel( Rectangle r ) {
		if( ! (mode == MODE.SOLVE || mode == MODE.EVAL) )
			return;
		
		Graphics g = buffer.getGraphics();
		for( int i=0; i<r.width; ++i )
			for( int j=0; j<r.height; ++j ) {
				if( solver.difficultylevel == null )
					continue;
				int		diff = solver.difficultylevel[r.x+i][r.y+j];
				if( diff <= 0 )
					continue;
				
				if( diff <= 1 )
					g.setColor(ColorLevel1);
				else if( diff <= 2 )
					g.setColor(ColorLevel2);
				else if( diff <= 3 )
					g.setColor(ColorLevel3);
				else if( diff <= 4 )				
					g.setColor(ColorLevel4);	
				else if( diff <= 5 )	
					g.setColor(ColorLevel5);
				else if( diff <= 7)	
					g.setColor(ColorLevel6);	
				else if( diff <= 30)					
					g.setColor(ColorLevel7);
				else
					g.setColor(ColorLevel99);

				int x = x0 + unit * (r.x+i);
				int y = y0 + unit * (r.y+j);
				g.fillRect(x+1, y+1, unit-1, unit-1);
			}
	}
	
	private void clearAreaList() {
		arealist = new ArrayList<Rectangle>();
	}
	
	private void addOneArea( Rectangle r ) {
		if( r == null )
			return;
		for(int i=arealist.size(); --i >=0 ; ){
			Rectangle a = arealist.get(i);
			if( a == null )
				continue;
			if( isOverLapped( r, a ) ) {
				arealist.remove(i);					// 重なった周囲の矩形は消す
			}
		}
		arealist.add(r);
	}
	
	// 分割のチェック　矩形があった場合、矩形の中の数字は、矩形の面積に一致していなければ取り除く
	void checkAreaList() {
		for(int i=arealist.size(); --i >=0 ; ){
			Rectangle a = arealist.get(i);
			if( a == null )
				continue;
			ArrayList<Point> pegs = getHintList(a);
			if( pegs.size()==1 ) {
				Point p = pegs.get(0);
				if( areatbl[p.x][p.y] == a.width*a.height ) {
					continue;
				}
			}
			arealist.remove(i);
		}
	}		
	
	private int howManyHint( Rectangle r ) {
		int cnt = 0;
		for( int i=0; i<r.width; ++i )
			for( int j=0; j<r.height; ++j )
				if( patterntbl[r.x+i][r.y+j] )
					++cnt;
		return cnt;
	}
	
	private Point whereisHint( Rectangle r ) {
		for( int i=0; i<r.width; ++i )
			for( int j=0; j<r.height; ++j )
				if( patterntbl[r.x+i][r.y+j] )
					return	new Point(r.x+i,r.y+j);
		return null;	
	}
	
	private ArrayList<Point> getHintList( Rectangle r ) {
		ArrayList<Point> pegs = new ArrayList<Point>();
		for( int i=0; i<r.width; ++i )
			for( int j=0; j<r.height; ++j )
				if( patterntbl[r.x+i][r.y+j] )	{
					pegs.add( new Point(r.x+i,r.y+j));
				}
		return	pegs;
	}
	
	private boolean checkHint( Rectangle r ) {
		int cnt = 0;
		int x = 0, y = 0;
		for( int i=0; i<r.width; ++i )
			for( int j=0; j<r.height; ++j )
				if( patterntbl[r.x+i][r.y+j] ) {
					++cnt;
					x = r.x+i;
					y = r.y+j;
				}
		if( cnt != 1 )
			return false;
		return areatbl[x][y] == r.width * r.height;
	}
	
	private boolean isOverLapped(Rectangle r1, Rectangle r2 ) {
		if( r1.x >= r2.x + r2.width || r2.x >= r1.x + r1.width )
			return false;
		if( r1.y >= r2.y + r2.height || r2.y >= r1.y + r1.height )
			return false;
		return true;
	}
	
	private void clearBuffer() {
		Graphics bufg = buffer.getGraphics();
		bufg.setColor(Color.white);
		bufg.fillRect(0, 0, getSize().width, getSize().height);		
	}
	
	private void drawSymErr(){
		Graphics bufg = buffer.getGraphics();
		bufg.setColor(Color.RED);	
		
		checkSymmetry();
		
		for( int i=0; i<bwidth; ++i )
			for( int j=0; j<bheight; ++j ) {
				if( symError[i][j] ) {
					int x = x0 + unit * i;
					int y = y0 + unit * j;
					bufg.drawLine(x , y, x+unit, y+unit);
					bufg.drawLine(x+unit , y, x, y+unit);
				}
			}
	}
	
	private void drawGrid(){
		Graphics bufg = buffer.getGraphics();
		bufg.setColor(Color.black);
		for(int i=0;i<=bwidth;i++){
			bufg.drawLine(x0 + i*unit, y0, x0 + i*unit , y0 + bheight*unit);
		}
		for(int i=0;i<=bheight;i++){
			bufg.drawLine(x0 , y0 + i*unit, x0 + bwidth*unit , y0 + i*unit);			
		}
	}
	
	private void setHintNumber() {
		int px = 0, py = 0;
		int cnt = 0;
		for( int i=0; i<rect.width; ++i )
			for( int j=0; j<rect.height; ++j ) {
				if( patterntbl[rect.x+i][rect.y+j] ) {
					px = rect.x+i;
					py = rect.y+j;
					++cnt;
				}
			}
		if(cnt==1) {
			if( mode == MODE.NUMBER ) {
				int	s = rect.width * rect.height;
				prenumtbl[px][py] = s > 1 ? s : 0;
			} else {
				areatbl[px][py] = rect.width * rect.height;
			}
		}
	}
	
	private void drawHints() {
		if( patterntbl == null )
			return;

		Graphics bufg = buffer.getGraphics();		
		bufg.setFont(numberFont);
		FontMetrics fm = bufg.getFontMetrics();
	
		
		for( int i=0; i<bwidth; ++i )
			for( int j=0; j<bheight; ++j ) {
				if( !patterntbl[i][j] )
					continue;
				Color circol = preshapetbl[i][j] ? hiddenPegColor : normalPegColor;
						
				bufg.setColor(circol);
				
				bufg.fillOval(x0+i*unit+1, y0+j*unit+1, unit-2, unit-2 );
				
				int	n = this.mode == MODE.NUMBER ? prenumtbl[i][j] : areatbl[i][j];
				if( n == 0 )
					continue;
				
				// 数字の表示
				bufg.setColor(Color.WHITE);				
				int xpad = x0+i*unit + (unit - fm.stringWidth(Integer.toString(n)))/2;
				int ypad = y0+j*unit + (unit - fm.getHeight())/2 + fm.getAscent();
				bufg.drawString(Integer.toString(n), xpad , ypad );
			}
	}

	// -------------------------------------------------------------------------
	// SymmetricProperty symmetry を元にチェックする。
	// -------------------------------------------------------------------------
	void checkSymmetry() {
		clearSymError();

		if( ! (symmetry.rotate2 || symmetry.vaxis || symmetry.haxis) )
			return;
		
		for(int i=0;i<bwidth;i++)
			for(int j=0;j<bheight;j++)
				if( ! checkSymmetry(i, j) )
					symError[i][j] = true;
	}
	
	boolean checkSymmetry(int x, int y) {
		Vector<Point> indice = new Vector<Point>();
		indice.add(new Point(x, y));
		symmetrize(indice);
		
		boolean	ors = false;
		boolean	ands = true;
		for (Point p : indice) {
			boolean s = patterntbl[p.x][p.y];
			ors |= s;
			ands &= s;
		}
		return	ors == ands;
	}
	
	private void expandPattern(int x, int y) {
		boolean state = patterntbl[x][y];

		Vector<Point> indice = new Vector<Point>();
		indice.add(new Point(x, y));
		symmetrize(indice);

		for ( Point p : indice ) {
			patterntbl[p.x][p.y]= state;
		}
	}
	
	private void symmetrize(Vector<Point> h) {
		for (int i = 0; i < h.size(); ++i) {
			Point p = h.get(i);
			Point q;
			
			if (symmetry.rotate2) {
				q = indexRotate2(p.x, p.y);
				if( Util.isOn(q,bwidth,bheight) )
					addIndex(h, q);
			}
			if (symmetry.vaxis) {
				q = indexVAxis(p.x, p.y);
				if( Util.isOn(q,bwidth,bheight) )
					addIndex(h, q);
			}
			if (symmetry.haxis) {
				q = indexHAxis(p.x, p.y);
				if( Util.isOn(q,bwidth,bheight) )
					addIndex(h, q);
			}
		}
	}
	
	// 対称性
	private Point indexRotate2(int x, int y) {
		return new Point(bwidth-1 - x, bheight-1 - y);
	}

	private Point indexVAxis(int x, int y) {
		return new Point(bwidth-1 - x, y);
	}
	
	private Point indexHAxis(int x, int y) {
		return new Point(x, bheight-1 - y);
	}
	
	private void addIndex(Vector<Point> h, Point p) {
		for (int i = 0; i < h.size(); ++i) {
			Point q = h.get(i);
			if (p.x == q.x && p.y == q.y)
				return;
		}
		h.add(p);
	}
	
	// ----------------------------- マウス --------------------------------
	
	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {
		int xi = toXIndex(e.getX());
		int yi = toYIndex(e.getY());
		if( 0>xi || xi>bwidth-1 || 0>yi || yi >bheight-1 )
			return;
		if(mode == MODE.SETTING || mode == MODE.CUTTING || mode == MODE.HIDDEN || mode == MODE.NUMBER ){
			if(	(e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0 ){
				changeCell(xi,yi);
				expandPattern(xi,yi);			// 対称性の展開
				repaint();
			}

			cursorMoveTo(xi, yi);
			if(	(e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0 ){
				rectStartTo(xi,yi);
			}
		}else if(mode == MODE.CUTTING) {
			cursorMoveTo(xi, yi);
			if(	(e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0 ){
				rectStartTo(xi,yi);
			}
		}else if(mode == MODE.PLAY){
				cursorMoveTo(xi, yi);
				rectStartTo(xi,yi);
				repaint();
		}

		arealistsaved = new ArrayList<Rectangle>(arealist);
		
		requestFocus();
	}
	
	public	void	changeCell( int x, int y ) {
		patterntbl[x][y] ^= true;	
		if( ! patterntbl[x][y] ) {
			areatbl[x][y] = 0;
		}
	}
	
	public void mouseReleased(MouseEvent e) {
		cursor = null;
		if( mode == MODE.CUTTING ) {
			checkAreaList();
		} else if( mode == MODE.HIDDEN ) {
			setHiddenArea();
			rect = null;
		}
		repaint();
	}
	public void mouseDragged(MouseEvent e){
		int xi = toXIndex(e.getX());
		int yi = toYIndex(e.getY());
		if( 0>xi || xi>bwidth-1 || 0>yi || yi > bheight-1 )
			return;		
			
		if( mode == MODE.CUTTING || mode == MODE.PLAY || mode == MODE.HIDDEN || mode == MODE.NUMBER ){		
			rectDraggedTo(xi,yi);
		}
		if( mode == MODE.CUTTING ) {
			checkAreaList();
		}
		repaint();
	}
	public void mouseMoved(MouseEvent e){}
	public void mouseWheelMoved(MouseWheelEvent e) {}
	
	// ----------------------------- キーボード --------------------------------
	
	public void keyPressed(KeyEvent e) {
		switch( e.getKeyCode() ) {
			case KeyEvent.VK_LEFT: cursorMove( -1, 0 );	return;
			case KeyEvent.VK_UP:   cursorMove( 0, -1 );	return;
			case KeyEvent.VK_RIGHT:cursorMove( +1, 0 );	return;
			case KeyEvent.VK_DOWN: cursorMove( 0, +1 );	return;
		}
		char c = e.getKeyChar();
		if(c == ' ' || (Character.isDigit(c))){		// 数値入力
			int v = c - '0';
			if(c==' ') v = 0 ; 
			if(mode == MODE.PLAY){
				int xi = cursor.x;
				int yi = cursor.y;
				getPlaytbl()[xi+1][yi+1] = v ;

				repaint();
			} else if( mode == MODE.SETTING || mode == MODE.CUTTING ) {
				int xi = cursor.x;
				int yi = cursor.y;
				if( v == 0 ) {
					if( patterntbl[xi][yi] )
						areatbl[xi][yi] = 0;
				}

				repaint();				
			}
		}
	}
	public void keyReleased(KeyEvent e) {}
	public void keyTyped(KeyEvent e) {}
	
	// -------------------------------------------------------------------------
	// コピー・ペースト
	// -------------------------------------------------------------------------	
	public void copybuffer() {
		if( rect == null )
			return;
		
		copybuffer = new boolean[rect.width][rect.height];
		copyarea   = new int[rect.width][rect.height];
		for( int i=0 ; i<rect.width; ++i )
			for( int j=0; j<rect.height; ++j ) {
				copybuffer[i][j] = patterntbl[rect.x+i][rect.y+j];
				copyarea[i][j]   = areatbl[rect.x+i][rect.y+j];
			}
	}
	
	public void pastebuffer() {
		if( copybuffer == null )
			return;
	
		for( int i=0 ; i<copybuffer.length; ++i )
			for( int j=0; j<copybuffer[0].length; ++j ) {
				int x = rectStart.x+i;
				int y = rectStart.y+j;
				patterntbl[x][y] = copybuffer[i][j];
				areatbl[x][y] = copyarea[i][j];
				
				expandPattern(x,y);			// 対称性の展開
			}
		
		repaint();
	}
	
	public void clearbuffer() {
		if( rect == null )
			return;
		
		copybuffer = null;
		for( int i=0 ; i<rect.width; ++i )
			for( int j=0; j<rect.height; ++j ) {
				patterntbl[rect.x+i][rect.y+j] = false;
				// 対称性の展開はしない
			}
		
		repaint();
	}
	
	// -------------------------------------------------------------------------
	// テーブルクリアなど
	// -------------------------------------------------------------------------
	void clearPlayTable(){
		for(int i=0;i<getPlaytbl().length;i++)
			Arrays.fill(getPlaytbl()[i] , 0);
		refresh();
	}
	void clearHintTable(){
		for(int i=0;i<bwidth;i++)
			for(int j=0;j<bheight;j++)
				getPattern()[i][j] = false;

		refresh();
	}

	int	countPegs() {
		int	cnt = 0;
		
		for(int i=0;i<bwidth;i++)
			for(int j=0;j<bheight;j++)
				if( patterntbl[i][j] )
					++cnt;

		return	cnt;
	}

	//---------------------------------------------------------------------
	//							解く
	//---------------------------------------------------------------------	
	private Difficulty getdifficultyfromGUI() {
		Difficulty sdiff = new Difficulty();
		if( difficultypanel != null ) {
			sdiff.initfixhalf = difficultypanel.getInitFixhalf();
			sdiff.pegfixhalf	= difficultypanel.getPegFixhalf();
			sdiff.celluniqpeg	= difficultypanel.getCelluniqpeg();
			sdiff.cellcommon	= difficultypanel.getCellcommon();
			sdiff.regiondivide	= difficultypanel.getRegiondivide();
		}
		return	sdiff;
	}
	
	void	solve( boolean quick ) {
		solver = new Solver(bwidth,bheight);
		Difficulty sdiff = getdifficultyfromGUI();	
		solvebylevel(sdiff, quick );
	}
	
	void	solvebyfulltech(boolean quick ) {
		solver = new Solver(bwidth,bheight);
		Difficulty usealldiff = new Difficulty(true);	
		solvebylevel(usealldiff, quick );
	}
	
	private void	solvebylevel( Difficulty usediff, boolean quick ) {		
		if( usediff != null) {
			solver.usedifficulty = usediff;
		}
		
		try {
			long starttime = System.currentTimeMillis();
			
			solver.eval( patterntbl, areatbl, quick );				// 解く
			
			long endtime = System.currentTimeMillis();
			double	runtime = (double)(endtime - starttime) / 1000;
			mainpanel.runtimeLabel.setText( String.format( "%8.3f", runtime ));
			
		} catch (ConflictExceptoin e) {
			e.print();
		}
		
		this.idtbl = solver.getCellIds();
		this.isrecs = solver.getIslandRectangles();

		Color buttoncolor = solver.isSuccess() ? Color.green : Color.RED;
		if( quick ) {
			mainpanel.solveButton.setBackground( buttoncolor );
		} else {
			mainpanel.evalButton.setBackground( buttoncolor );
		}

		int	sz = solver.answersequence.size();
		if( difficultypanel != null ) {			
			difficultypanel.historypanel.setslidersize( sz );
			replayposition = sz;
		}
		
		solver.applySequence(sz);	
		isrecs = solver.getSequenceRectangles();
		
		setDifficultypoint( solver.getDiffPoint() );
	}
	
	void	setcut() {
		arealist = isrecs!=null ? isrecs : new ArrayList<Rectangle>();
		isrecs = null;
		mainpanel.solveButton.setBackground(null);
		mainpanel.evalButton.setBackground(null);		
	}
	
	//---------------------------------------------------------------------
	//							Clear
	//---------------------------------------------------------------------	
	void	clearBoard() {
		if( mode == MODE.SETTING ) {
			arealist = new ArrayList<Rectangle>();	
			clearAllArea();	// Hintの数字（面積）をクリア
			patterntbl = new boolean[bwidth][bheight];
			isrecs = null;
			mainpanel.solveButton.setBackground(null);
			mainpanel.evalButton.setBackground(null);
			
		} else if( mode == MODE.CUTTING ) {
			if( isrecs != null ) {
				isrecs = null;
				mainpanel.solveButton.setBackground(null);
				mainpanel.evalButton.setBackground(null);				
			} else {
				arealist = new ArrayList<Rectangle>();	
				areatbl = new int[bwidth][bheight];
			}
			
		} else if( mode == MODE.PLAY ) {
			arealist = new ArrayList<Rectangle>();					
		} else if( mode == MODE.SOLVE || mode == MODE.EVAL ){
			isrecs = null;
			arealist = new ArrayList<Rectangle>();					
		}
	}
	
	// Hintの数字（面積）をクリア
	void	clearAllArea() {
		areatbl = new int[bwidth][bheight];
	}
	
	int		replayposition;
	void	setReplayposition( int n ) {
		replayposition = n;
	}
	void	replay( int n ) {
		if( replayposition == n )
			return;	
		replayposition = n;	
		
		idtbl = new int [bwidth][bheight];
		isrecs = new ArrayList<Rectangle>();
		solver.applySequence(n);	
		isrecs = solver.getSequenceRectangles();
		refresh();
	}

	// ------------------------- File -----------------------------

	// ペグのパターン
	String currentPatternString()
	{
		StringBuffer buf = new StringBuffer();
		buf.append("\n");
		for(int j=0;j<bheight;j++){
			for(int i=0;i<bwidth;i++){
				buf.append(patterntbl[i][j]?"P ":". ");
			}
			buf.append("\n");
		}
		return buf.toString();
	}
	
	String currentPattern2String()
	{
		StringBuffer buf = new StringBuffer();
		buf.append("\n");
		for(int j=0;j<bheight;j++){
			for(int i=0;i<bwidth;i++){
				buf.append(preshapetbl[i][j]?"Q ":". ");
			}
			buf.append("\n");
		}
		return buf.toString();
	}
	
	String currentPattern3String()
	{
		StringBuffer buf = new StringBuffer();
		buf.append("\n");
		for(int j=0;j<bheight;j++){
			for(int i=0;i<bwidth;i++){
				buf.append(prenumtbl[i][j]+" ");
			}
			buf.append("\n");
		}
		return buf.toString();
	}
	
	// ペグの面積データ
	String currentAreaString()
	{
		StringBuffer buf = new StringBuffer();
		buf.append("\n");
		for(int j=0;j<bheight;j++){
			for(int i=0;i<bwidth;i++){
				int n = patterntbl[i][j]?areatbl[i][j]:0;
				buf.append(" " + n);
			}
			buf.append("\n");
		}
		return buf.toString();
	}
	
	int	getBoardWidth()  {	return	bwidth;	}
	int	getBoardHeight() {	return bheight;	}
	
	int getUnit() {
		return	unit;
	}
	public void setPattern(boolean[][] pattern) {
		this.patterntbl = pattern;
	}
	public boolean[][] getPattern() {
		return patterntbl;
	}
	
	public	boolean	getPattern(int x, int y ) {
		return	patterntbl[x][y];
	}
	
	public void setPattern2(boolean[][] pattern2) {
		this.preshapetbl = pattern2;
	}	
	
	public void setArea(int[][] area) {
		this.areatbl = area;
	}
	public int[][] getArea() {
		return areatbl;
	}	
	
	public ArrayList<Rectangle> getCheckedArealist() {
		return	getCheckedArealist(arealist);
	}
	
	public ArrayList<Rectangle> getCheckedArealist(ArrayList<Rectangle> rlist ) {
		ArrayList<Rectangle>	ar = new ArrayList<Rectangle>();
		for( Rectangle r : rlist ) {
			if( checkHint(r) )
				ar.add(r);
		}
		
		return	ar;
	}
	
	public void setPlaytbl(int[][] playtbl) {
		this.playtbl = playtbl;
	}
	public int[][] getPlaytbl() {
		return playtbl;
	}
	public void clearPlaytbl() {
		for( int i=0; i<playtbl.length; ++i )
			for( int j=0; j<playtbl[i].length; ++j )
				playtbl[i][j] = 0;
	}
	
	void clearSymError() {
		for(int i=0;i<bwidth;++i)
			for( int j=0; j<bheight; ++j )
				symError[i][j] = false;
	}	
	
	//---------------------------------------------------------------------
	//							形状カウント
	//---------------------------------------------------------------------	
	ArrayList<ShapeCount>		shapelist;
	
	public String	getShapeCounts() {
		makeShapeList( getCurrentRectList() );
		return getShapeListString();
	}
	
	class ShapeCount {
		int		area;			//　面積
		int		longside;		//	長辺
		int		shortside;		//	短辺
		int		count;			//	個数
	}
	
	public	void	makeShapeList( ArrayList<Rectangle> recs ) {
		if( recs == null )
			return;
		
		shapelist = null;
		for( Rectangle rec : recs ) {
			addShape(rec);
		}
	}
	
	private	void	addShape( Rectangle rec) {
		if( shapelist == null ) {
			shapelist = new ArrayList<ShapeCount>();
		}
		
		int	lside	= Math.max( rec.width, rec.height );
		int sside	= Math.min( rec.width, rec.height );
		for( ShapeCount shc : shapelist) {
			if( shc.longside == lside && shc.shortside == sside ) {
				++shc.count;
				return;
			}
		}
		ShapeCount	sh = new ShapeCount();
		sh.area = rec.width*rec.height;
		sh.longside = lside;
		sh.shortside = sside;
		sh.count = 1;

		shapelist.add(sh);
	}
	
	// 面積順に印刷
	private	String	getShapeListString() {
		int		minarea = 0;
		int		maxarea = 0;
		if( shapelist == null )
			return null;
		
		for( ShapeCount sc : shapelist ) {
			if( minarea == 0 ) {
				minarea = sc.area;
			} else if( minarea > sc.area ) {
				minarea = sc.area;
			}
			if( maxarea == 0 ) {
				maxarea = sc.area;
			} else if( maxarea < sc.area ) {
				maxarea = sc.area;
			}
		}

		StringBuffer	strbuf = new StringBuffer();
		
		for( int s = minarea; s<=maxarea; ++s ) {
			ArrayList<ShapeCount> samesize = new ArrayList<ShapeCount>();
			for( ShapeCount sc : shapelist ) {
				if( sc.area == s )
					samesize.add(sc);
			}
			ArrayList<ShapeCount> work = new ArrayList<ShapeCount>();
			for( int sh=1;samesize.size()>0; ++sh ) {
				for( ShapeCount w : samesize ) {
					if( w.shortside == sh ) {
						samesize.remove(w);
						work.add(w);
						break;
					}
				}
			}
			samesize = work;
			if( samesize.size() == 0 )
				continue;
			
			int	cnt = 0;
			for( ShapeCount sh : samesize )
				cnt += sh.count;
			
			strbuf.append(String.format(" %3d  %3d  ", s, cnt ));
			for( ShapeCount sh : samesize ) {
				strbuf.append(String.format( " %dx%d=%d", sh.longside, sh.shortside, sh.count));
			}
			strbuf.append(String.format("%n"));
		}
		return	strbuf.toString();
	}
	
	public	ArrayList<Rectangle>	shapeerrorrecs( ArrayList<Rectangle> reclist ) {
		return	difficultypanel.getshapeerrorrecs(reclist);
	}
	
	//------------------------------------------------------------------------------
	//	塊を数える
	//------------------------------------------------------------------------------
	int		getlumpcount() {
		double	threshold = ga.getuneventhreshold();
		int		step = ga.getunevenstep();
		
		return	getlumpcount( threshold, step, true );
	}
	
	int		getlumpcount( double threshold, int maxstep, boolean prt ) {	
		int[][] difftable = getDifficultytable();
		int	n = 0;
		int	sum = 0;
		
		for( int x=0; x<bwidth; ++x )
			for( int y=0; y<bheight; ++y ) {
				int	diff = difftable[x][y];	
				if( diff < 1 )
					continue;
				++n;
				sum += diff;
			}
		
		double	average = n > 0 ? (double)sum/n : 1.0;
		if( prt ) {
			System.out.println("average="+average);
			printave(difftable);
		}

		if( threshold > 0.0 )
			average = threshold;

		if( prt ) {
			System.out.println("average="+average);
		}
		
		boolean[][]	bintable = new boolean[bwidth+2][bheight+2];
		for( int x=1; x<=bwidth; ++x )
			for( int y=1; y<=bheight; ++y ) {
				bintable[x][y] = difftable[x-1][y-1] < average;
			}
		
		boolean[][]	edgetable = new boolean[bwidth+2][bheight+2];

		for( int x=0; x<bwidth+2; ++x )
			for( int y=0; y<bheight+2; ++y ) {
				if( x==0 || x==bwidth+1 || y==0 || y==bheight+1 ) {
					continue;
				}
				edgetable[x][y] = bintable[x][y];
			}
		for( int x=1; x<bwidth+1; ++x ) {
			edgetable[x][0        ] = bintable[x][1];
			edgetable[x][bheight+1] = bintable[x][bheight];
		}
		for( int y=1; y<bheight+1; ++y ) {
			edgetable[0       ][y] = bintable[1][y];
			edgetable[bwidth+1][y] = bintable[bwidth][y];
		}
			
		edgetable[0     ][0      ] = true;
		edgetable[0     ][bheight] = true;
		edgetable[bwidth][0      ] = true;
		edgetable[bwidth][bheight] = true;

		boolean[][]	previous = edgetable;

		if( prt ) {
			System.out.println("expand 0"); printbintable(previous);
		}
		
		for( int i=0; i<maxstep; ++i ) {
			boolean[][] next = new boolean[bwidth+2][bheight+2];
			for( int x=0; x<bwidth+2; ++x )
				for( int y=0; y<bheight+2; ++y ) {
					next[x][y] = previous[x][y];
				}
			next[0][0] = previous[1][1];
			next[0][bheight+1] = previous[1][bheight];
			next[bwidth+1][0] = previous[bwidth][1];
			next[bwidth+1][bheight+1] = previous[bwidth][bheight];

			for( int x=1; x<bwidth+1; ++x )
				for( int y=1; y<bheight+1; ++y ) {
					boolean p = previous[x][y];
					if( ! p )
						continue;
					next[x][y] = previous[x-1][y] && previous[x+1][y] && previous[x][y-1] && previous[x][y+1];
				}
			previous = next;
			if( prt ) {
				System.out.println("expand"+(i+1)); printbintable(previous);
			}
		}
		
		int	val = 0;
		for( int x=1; x<=bwidth; ++x )
			for( int y=1; y<=bheight; ++y )
				if( previous[x][y] )
					++val;
				
		return	val;
	}
	
	private void printbintable( boolean[][] bintable) {
		int	count = 0;
		for( int y=1; y<=bheight; ++y ) {
			for( int x=1; x<=bwidth; ++x ) {
				System.out.printf(bintable[x][y]? " X" : " _");
				count += bintable[x][y] ? 1 : 0;
			}
			System.out.println();
		}
		System.out.printf("X counts = %d%n", count );
	}
		
	private	void	printave( int[][] aves ) {
		System.out.println("---- average table ");
		for( int y=0; y<bheight; ++y ) {
			for( int x=0; x<bwidth; ++x )
				System.out.printf("%3d", aves[x][y] );
			System.out.println();
		}
	}
}
