package numarea;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Stack;

public class Solver {
	int 						width, height;
	boolean[][] 				hinttbl;
	int[][]						areatbl;

	private Cell[][]			cellboard;			// Cell盤面
	private ArrayList<Peg>		allpegs;			// Pegs
	private ArrayList<Island>	allislands;			// 島
	private Fixable				nextfixable;		// 次に決める場所

	private int					islandcnt;			//   id
	private	boolean				success;			// 	成功
	public boolean isSuccess()	{	return	success;	}

	private	double	currentdiffpoint;
	private	double	totaldiffpoint;					// 難易度ポイント

	public	Fixable				currentstep;
	public	ArrayList<Fixable> 	answersequence;		// 解き順
	public	ArrayList<Rectangle>	answerrecs;
	private	boolean				isreplay;			// 再生時にtrue

	final	int[][]	direction = { {1,0}, {0,1}, {-1,0}, {0,-1} };

	//----------------------- 難易度関連データ -----------------------
	int[][]			difficultylevel;		// 難易度レベル

	// 決定タイプ（手筋）
	final	static	int		INIT_FIX_FULL = 2;
	final	static	int		INIT_FIX_HALF	= 4;
	final	static	int		PEG_FIX_FULL	= 3;
	final	static	int		PEG_FIX_HALF	= 5;
	final	static	int		CELL_UNIQPEG_PLUG = 1;
	final	static	int		CELL_UNIQPEG_FULL	= 8;	
	final	static	int		CELL_UNIQPEG	= 6;	
	final	static	int		CELL_COMMON	= 7;
	final	static	int		REGION_DIVIDE = 9;
	final	static	int		UNKNOW_FILL	= 99;

	// 難易度の基礎ポイント
	final	static	int		POINT_PEG_FIX_FULL = 2;
	final	static	int		POINT_PEG_FIX_HALF = 3;
	final	static	int		POINT_CELL_UNIQPEG_FULL = 4;
	final	static	int		POINT_CELL_UNIQPEG = 10;
	final	static	int		POINT_CELL_UNIQPEG_SP1A = 1;
	final	static	int		POINT_CELL_UNIQPEG_SP1B = 2;
	final	static	int		POINT_CELL_UNIQPEG_SP2 = 15;	

	final	static	int		POINT_CELL_COMMON	= 20;
	final	static	int		POINT_REGION_DIVIDE_BASE = 10;
	final	static	int		POINT_REGION_DIVIDE_STEP = 10;	

	public Difficulty		usedifficulty = new Difficulty();

	//--------------------------------------------------------------------------
	//		Solver
	//--------------------------------------------------------------------------
	Solver(int w, int h) {
		this.width = w;
		this.height = h;
		answerrecs = new ArrayList<Rectangle>();	
	}

	void eval( boolean[][] hint, int[][] area, boolean quick ) throws ConflictExceptoin {		
		hinttbl = hint;
		areatbl = area;
		answersequence = new ArrayList<Fixable>();	
		answerrecs = new ArrayList<Rectangle>();	

		success = false;	
		isreplay = false;

		difficultylevel = new int[width][height];	
		totaldiffpoint = 0.0;

		solveInitSetup();					// 初期設定
		totaldiffpoint += initialPossiblePoint();

		for(;;){
			if( isAllFinished() )
				break;

			nextfixable = null;
			currentdiffpoint = 0;

			checkPegs();						// ペグ矩形チェック

			for( int i=0; i<1; ++i ) {
				if( quick && nextfixable!=null )
					break;

				if( usedifficulty.deadend ) {
					checkCells( true, quick );						// 凸
				}
				if( quick && nextfixable!=null )
					break;

				if( usedifficulty.celluniqpeg ) {
					checkCells( false, quick );						// セル到達チェック
				}
				if( quick && nextfixable!=null )
					break;

				if( usedifficulty.regiondivide && nextfixable == null ) {
					checkRegionBreak( quick );		// 領域分割
				}
				if( quick && nextfixable!=null )
					break;

				if( usedifficulty.cellcommon ) {
					checkCellCommonIslands( quick );	// セルから共通の島を見つける
				}
			}

			if( nextfixable == null)	 {		// 何も見つからなかったので終了
				break;
			}

			applyOne(nextfixable);

			currentdiffpoint += +nextfixable.level;		// level値を基礎点数として加える

			currentstep = getCurrentStep();
			currentstep.diff = currentdiffpoint;
			totaldiffpoint += currentdiffpoint;
		}

		fillRestLevels();
		totaldiffpoint += hardpoints();
		checkSuccess();

		if( success ) {
			for( Peg p : allpegs ) {
				if( !p.finished )
					continue;
				answerrecs.add(p.island.rec);
			}
		}

		// 難易度調整
		double npeg = countPegs();
		totaldiffpoint *= Math.E/Math.log(width*height/npeg);
	}

	// 初期化時に決まる難易度ポイントの計算
	int	hardpoints() {
		int pnt = 0;

		for( int j=0; j<height; ++j ) {
			for( int i=0; i<width; ++i ) {
				int p = difficultylevel[i][j];
				if( p>3 )
					pnt += p*p;
			}	
		}
		return	pnt;
	}

	//--------------------------------------------------------------------------
	//		初期化
	//--------------------------------------------------------------------------	
	void solveInitSetup() throws ConflictExceptoin {
		cellboard = new Cell[width][height];
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				cellboard[i][j] = new Cell(i,j);
			}

		allpegs = new ArrayList<Peg>();
		allislands = new ArrayList<Island>();
		islandcnt = 0;

		// ペグの初期生成　カバー可能性だけを生成する
		int id = 0;
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				if( hinttbl[i][j] ) {			// ペグのときだけ処理をする
					++id;
					difficultylevel[i][j] = -1;
					Peg pg  = new Peg(id,i,j,areatbl[i][j]);
					allpegs.add(pg);
					Cell cell = cellboard[i][j];
					cell.peg = pg;
					cell.ispeg = true;
					pg.initPossibles();			// 矩形候補を加える
					pg.initEval();

				}
			}

		// Cellの所属ペグ候補を生成する。
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				Cell cell = cellboard[i][j];
				if( cell.ispeg )					// ペグの場合は無視
					continue;
				cell.initBelong();
			}
	}

	private	int		initialPossiblePoint() {
		int	point = 0;
		for( Peg p : allpegs ) {
			if( p.possibles != null ) {
				point += p.possibles.size();
			}
		}
		return	point;
	}

	//--------------------------------------------------------------------------
	//	待ち行列の取り出し、実行
	//--------------------------------------------------------------------------
	private	void	applyOne( Fixable f ) {
		switch( f.type ) {
		case INIT_FIX_FULL:
		case INIT_FIX_HALF:

		case PEG_FIX_FULL:
		case PEG_FIX_HALF:
			f.peg.apply( f.type, f.level, f.rec );		
			break;

		case CELL_UNIQPEG_PLUG:			
		case CELL_UNIQPEG_FULL:			
		case CELL_UNIQPEG:
			f.cell.applyBelong( f.type, f.level );		
			break;

		case CELL_COMMON:
			f.cell.applyCommomPossibles(f.type, f.level, f.rec );

			break;

		case REGION_DIVIDE:
			f.peg.applyRemoveRec( f.type, f.level, f.rec );		

			break;

		default:
		}		
	}

	//--------------------------------------------------------------------------
	//	Step n までの履歴シーケンスだけを実行する
	//--------------------------------------------------------------------------
	private	ArrayList<Rectangle>	sequencerecs;

	void	addAnswerSequence( Fixable fnew ) {
		for( Fixable f : answersequence ) {
			if( f.rec == null || fnew.rec == null )
				continue;
			Rectangle com = f.rec.intersection(fnew.rec);
			if( fnew.rec.equals(com) )
				return;	
		}
		answersequence.add(fnew);
	}

	public	void	applySequence( int n ) {
		isreplay = true;
		sequencerecs = new ArrayList<Rectangle>();

		difficultylevel = new int[width][height];	
		try {
			solveInitSetup();	
		} catch(ConflictExceptoin e){}

		for( int i=0; i<n; ++i ) {
			if( i >= answersequence.size() )
				break;
			Fixable f = answersequence.get(i);
			applyOneHistory(f);
			sequencerecs.add( f.rec );
		}
	}

	ArrayList<Rectangle>	getSequenceRectangles() {
		return	sequencerecs;
	}

	private	void	applyOneHistory( Fixable f ) {
		if( f.rec != null )
			setLevelRectangle( f.rec, f.type, f.level );
	}

	//--------------------------------------------------------------------------
	// 全ペグについて、調べる			PEG_FIX_FULL, PEG_FIX_HALF
	//--------------------------------------------------------------------------
	private void	checkPegs() throws ConflictExceptoin {
		for( Peg pg : allpegs ) {
			if( pg.finished )
				continue;			
			pg.checkPossibles();
		}
	}

	//--------------------------------------------------------------------------
	// 全Cellについて調べる			CELL_PEG
	//--------------------------------------------------------------------------
	private void checkCells( boolean deadend, boolean quick ) throws ConflictExceptoin {
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				Cell c = cellboard[i][j];
				c.checkBelong(deadend);
				if( quick && nextfixable!=null )
					return;
			}
	}

	// 
	//--------------------------------------------------------------------------
	// 	セルから共通の島を見つける		CELL_COMMON
	//--------------------------------------------------------------------------
	private void	checkCellCommonIslands( boolean quick ) {
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				// Cellの候補のペグすべての矩形の共通部分を探す。
				Cell c = cellboard[i][j];
				if( c.peg == null && c.island == null ) {	// 所属ペグが決まっていない時に調べる
					c.checkCommomPossibles();
				}
				if( quick && nextfixable!=null )
					return;
			}		
	}

	//--------------------------------------------------------------------------
	// 全ての領域をチェックする
	//--------------------------------------------------------------------------
	private void	checkRegionBreak( boolean quick ) {
		Region reg = new Region();
		//　全ペグの全候補矩形について調べる
		for( Peg pg : allpegs ) {
			if( pg.num == 0 )
				continue;
			if( pg.finished )
				continue;
			for( Rectangle r : pg.possibles) {
				boolean possible = reg.checkRectangle( r );	
				if( !possible ) {		// その置き方は不可能
					int pnt = POINT_REGION_DIVIDE_BASE + POINT_REGION_DIVIDE_STEP * reg.minpegcnterr;
					bidForNext( new Fixable( REGION_DIVIDE, pnt, pg, r ) );
				}
			}
			if( quick && nextfixable!=null )
				return;
		}
	}

	//--------------------------------------------------------------------------
	// 盤面上の全てのペグの個数を数えます。
	private	int		countPegs() {
		int cnt = 0;
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j )
				if( hinttbl[i][j] )
					++cnt;
		return cnt;
	}

	// 全cellに対して、決定したidを求める。半決定の場合、負のid
	int[][] getCellIds() {
		int[][] ids = new int[width][height];
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				Peg p = cellboard[i][j].peg;
				if( p!= null )
					ids[i][j] = p.finished ? p.id : -p.id;
			}

		return	ids;
	}

	ArrayList<Rectangle>	getIslandRectangles() {
		ArrayList<Rectangle>	recs = new ArrayList<Rectangle>();
		for( Island il : allislands ) {
			recs.add( il.rec );
		}
		return	recs;
	}

	// 成功かどうかの判定		全ペグ、全マスの調査
	private void checkSuccess() {
		success = false;
		for( Peg p : allpegs ) {
			if( ! p.finished ) {
				return;
			}
		}

		for( int i=0; i<width; ++i ) {
			for( int j=0; j<height; ++j ) {			
				Cell c = cellboard[i][j];	
				if( c.peg==null || c.peg.id<=0 ) {
					return;
				}
			}
		}

		success = true;
	}

	void fillRestLevels() {
		for( int i=0; i<width; ++i )
			for( int j=0; j<height; ++j ) {
				Cell c = cellboard[i][j];
				if( c.island!=null ) {
					if( difficultylevel[i][j] == 0 )
						difficultylevel[i][j] = UNKNOW_FILL;
				}
			}
	}

	//	矩形内で、まだレベルが未定なら、指定したレベルを設定する
	void	setLevelRectangle( Rectangle rec, int tp, int lev ) {

		if (rec != null) {
			for (int i = 0; i < rec.width; ++i)
				for (int j = 0; j < rec.height; ++j) {
					int x = rec.x + i;
					int y = rec.y + j;
					if (difficultylevel[x][y] <= 0) {
						difficultylevel[x][y] = lev;
					}
				}
		}

		if( ! isreplay ) {
			addAnswerSequence( new Fixable( tp, lev, rec ) );			
		}
	}

	boolean	isOnBoard( Rectangle r ) {
		if( r.x < 0 || r.y < 0 )
			return	false;
		if( r.x + r.width - 1 >= this.width || r.y + r.height - 1 >= this.height )
			return	false;
		return	true;
	}

	boolean	isOnBoard( Point p ) {
		if( p.x < 0 || p.y < 0 )
			return	false;
		if( p.x >= this.width || p.y >= this.height )
			return	false;
		return	true;
	}

	boolean isAllFinished() {
		for( Peg pg : allpegs ) {
			if( !pg.finished )
				return	false;		
		}
		return	true;
	}

	//==========================================================================
	//	ペグ
	//==========================================================================
	class	Peg {						
		int						id;				// id  1..
		Point					pos;			// 位置
		int						num;			// ペグの数値（面積）
		ArrayList<Rectangle>	possibles;		// 可能な矩形候補の全リスト
		Island					island;		// 確定領域
		boolean				finished;		// 完成
		boolean				wait;			// 処理ありだけれど、待ち状態

		Peg( int i, int x, int y, int pn ) {
			id = i;
			pos = new Point(x,y);
			num = pn;
			possibles = null;
			island = null;
			finished = false;
			wait	= false;
		}

		// 可能な矩形パターンを集める
		void initPossibles() {
			if( num == 0 )
				return;
			ArrayList<Dimension> combinations = Util.getCombinations(num);
			possibles = new ArrayList<Rectangle>();
			for( Dimension comb : combinations ) {
				for( int i=0; i<comb.width; ++i )
					for( int j=0; j<comb.height; ++j ) {
						Rectangle rec = new Rectangle(pos.x-i,pos.y-j,comb.width,comb.height);
						if( ! isOnBoard(rec) )
							continue;
						if( ! anyOtherPegs( rec ) ) {
							possibles.add(rec);
						}
					}
			}
		}

		// 初期状態で決まるかどうかを判断する
		//	0: 何も決まらない
		//	1: 初期完全確定 INIT_FIX_FULL
		//	2: 初期部分確定 INIT_FIX_HALF
		void	initEval() throws ConflictExceptoin {
			if( num == 0 || possibles == null )
				return;
			Rectangle com = Util.getCommonRectangle(possibles);
			if( com==null ) {
				String s = String.format("Peg.initEval no common rectangle area.\nPeg %d(%d,%d)", num, pos.x, pos.y );
				throw new ConflictExceptoin(s);
			}
			int	sz = com.width * com.height;		// 共通部分の面積
			if( sz == 1 )
				return;
			int t = possibles.size() == 1 ? INIT_FIX_FULL : INIT_FIX_HALF;
			int lev = t + possibles.size() - 1;

			if( ! usedifficulty.initfixhalf ) {
				if( lev == INIT_FIX_HALF ) {
					return;
				}
			}

			bidForNext( new Fixable(t, lev, this, com ));				// type, level
		}

		private boolean anyOtherPegs( Rectangle rec ) {
			for( int i=0; i<rec.width; ++i )
				for( int j=0; j<rec.height; ++j ) {
					int px = rec.x + i;
					int py = rec.y + j;
					if( ! hinttbl[px][py] )					// ペグがなければ無視
						continue;
					if( px == pos.x && py == pos.y )		// 自分のペグは無視する
						continue;
					return	true;								// 他のペグが存在した
				}
			return	false;
		}

		// 可能性矩形について、本当に可能か調べる。
		// ダメな矩形は取り除き、島の形を調べる。
		// 島の形が変われば、changed=true とする。
		boolean checkPossibles() throws ConflictExceptoin {
			if( possibles == null ) {		// 長方形候補リストが与えられていない場合、無視する。
				return false;
			}
			for( int i=possibles.size(); --i>=0; ) {
				Rectangle r = possibles.get(i);
				if( checkInterfered( r ) )
					possibles.remove(i);
			}
			Rectangle	newcom = Util.getCommonRectangle(possibles);
			if( newcom == null ) {		// 可能矩形の共通部分が存在しない　→　破綻！
				String s = String.format("checkPossibles no common rectangle area.\nPeg %d(%d,%d)", num, pos.x, pos.y );
				throw new ConflictExceptoin(s);
			}
			if( island != null ) {
				Rectangle	oldcom = island.rec;
				if( newcom.equals(oldcom) )
					return false;
			}

			int sz = newcom.width * newcom.height;
			if( sz == 1 )
				return false;

			int	t, lev;
			if( newcom.width * newcom.height == num ) {
				t = PEG_FIX_FULL;
				lev = POINT_PEG_FIX_FULL;
			} else {
				t = PEG_FIX_HALF;
				lev = POINT_PEG_FIX_HALF + (num - newcom.width * newcom.height);
			}

			if( ! usedifficulty.pegfixhalf ) {
				if( t == PEG_FIX_HALF )
					return false;
			}

			bidForNext( new Fixable( t, lev, this, newcom ));							// type, level

			return	true;
		}

		// Peg の共通領域を rec に変更し、影響を調整する。
		void apply( int tp, int lev, Rectangle rec ) {
			if( finished )
				return;

			// 島の形が変わった。	
			if( island == null ) {
				island = new Island( rec, this );
				allislands.add( island );
			} else {
				island.expandTo( rec );
			}
			setLevelRectangle( rec, tp, lev );

			if( possibles.size() == 1 ) {
				finished = true;
				possibles = null;
			}
		}

		void applyRemoveRec( int tp, int lev, Rectangle rec ) {
			if( finished )
				return;

			for( Rectangle r : possibles ) {
				if( rec.equals(r) ) {
					possibles.remove(r);
					break;
				}
			}

			setLevelRectangle( null, tp, lev );
		}

		// CellがこのPegに属することが分かったときに、可能性を調べ直す。
		// ダメな矩形は取り除き、島の形を調べる。
		// 島の形が変われば、changed=true とする。
		void checkPossiblesCell( int tp, int lev, Cell c ) {
			if( possibles == null ) {		// 面積が与えられていないばあい、無視する。
				return;
			}
			for( int i=possibles.size(); --i>=0; ) {
				Rectangle r = possibles.get(i);
				if( ! r.contains( c.pos ) )	 {			
					possibles.remove(i);
					continue;
				}
				if( checkInterfered( r ) ) {
					possibles.remove(i);
				}
			}
			Rectangle	newcom = Util.getCommonRectangle(possibles);
			if( newcom == null )
				return;							// 破綻！

			if( island != null ) {
				Rectangle	oldcom = island.rec;
				if( newcom.equals(oldcom) )
					return;
				island.expandTo( newcom );
			} else {
				island = new Island( newcom, this );
				allislands.add( island );
			}

			// 島の形が変わった。
			setLevelRectangle( newcom, tp, lev );	

			if( possibles.size() == 1 ) {			
				finished = true;
				possibles = null;
			}
		}

		// ペグの候補矩形が、実際に可能かどうか調べる
		//　邪魔されていて、不可能だったらtrue
		private boolean checkInterfered( Rectangle rec ) {
			for( int i=0; i<rec.width; ++i )
				for( int j=0; j<rec.height; ++j ) {
					int px = rec.x + i;
					int py = rec.y + j;
					if( px == pos.x && py == pos.y )		// 自分のペグは無視する
						continue;
					if( hinttbl[px][py] )					// 他のペグが存在した
						return	true;
					Cell c = cellboard[px][py];
					if( c.peg == null ) {
						if( c.island == null )
							continue;
						// 島であった
						Rectangle uni = Util.getCommonRectangle( rec, c.island.rec );
						if( uni == null )						// 島が重ならない
							continue;
						if( uni.equals( c.island.rec ))		// 島が完全に含まれる
							continue;
						return	true;							// 島が部分的に引っかかる
					}
					if( c.peg != this	 )
						return	true;							// 他のペグに属するcellが見つかった
				}
			return	false;		
		}

		// 指定のセルを含む候補矩形の個数
		int	countPossibles( Cell c ) {
			int	cnt = 0;
			if( possibles != null ) {
				for( Rectangle r : possibles ) {
					if( r.contains(c.pos) ) {
						++cnt;
					}
				}
			}
			return	cnt;
		}
	}

	//==========================================================================
	// 									島
	//==========================================================================
	class	Island { 
		int						id;	
		Rectangle				rec;			// 島
		int						size;			// 面積
		boolean				finished;		// 完成
		Peg						peg;			// 決定peg (所属)

		Island( Rectangle rc, Peg pg ) {
			this.id  = ++islandcnt;
			this.rec = rc;
			this.peg = pg;			
			this.finished = false;
			if( rec != null ) {
				setInsideCells();
			}
		}

		private void expandTo( Rectangle com ) {
			rec = com;
			setInsideCells();
		}

		// 島内部のセルの処理
		private void	setInsideCells() {
			size = rec.width * rec.height;
			for( int i=0; i<rec.width; ++i )
				for( int j=0; j<rec.height; ++j ) {
					int x = rec.x+i;
					int y = rec.y+j;
					Cell c = cellboard[x][y];
					c.peg = this.peg;
					if( c.peg != null ) {	// ペグが決定なら、belongsは不要
						c.belongs = null;	
					}
					if( c.island != null && c.island != this ) {
						allislands.remove(c.island);
					}
					c.island = this;
				}		
		}
	}

	//==========================================================================
	// 									セル
	//==========================================================================
	class	Cell {						
		Point					pos;			// マスの位置
		boolean				ispeg;			// ペグかどうか
		ArrayList<Peg>		belongs;		// 候補peg
		Peg						peg;			// 決定peg (所属)
		Island					island;		// 島

		Cell( int x, int y ) {
			this.pos = new Point(x,y);
			this.ispeg = false;
			this.belongs = new ArrayList<Peg>();
			this.peg = null;
			this.island = null;
		}

		private void fixPeg( Peg p, Island land ) {
			this.peg = p;
			this.belongs = null;
			this.island = land;
		}

		// 可能なペグを集める		
		private void initBelong() {
			if( peg != null )
				return;
			for( Peg p : allpegs ) {
				if( p.num == 0 ) {						// 数字未定ペグの場合
					Rectangle r = Util.makeRectangle( this.pos, p.pos );
					boolean find = false;
					for( Peg q : allpegs ) {
						if( p == q )
							continue;
						if( r.contains( q.pos ) ) {					
							find = true;
							break;
						}
					}
					if( !find )
						belongs.add(p);
				} else {									// 数字のあるペグの場合
					boolean inc = false;
					if( p.possibles == null )
						continue;
					for( Rectangle r : p.possibles ) {
						if( r.contains( pos ) ) {
							inc = true;
							break;
						}
					}
					if( inc )
						belongs.add(p);
				}
			}	
		}

		//--------------------------------------------------------------------------
		//	待ち行列、CELL_PEG の検査
		//	このセルに届く特定のペグの候補矩形の共通矩形は島になる。
		//--------------------------------------------------------------------------
		private void checkBelong(boolean	deadend) throws ConflictExceptoin {
			if( ispeg || peg != null || belongs==null )
				return;

			for( int i=belongs.size(); --i >=0 ; ) {			// 所属ペグにまだ所属しているか調べる
				Peg p = belongs.get(i);
				Rectangle r = Util.makeRectangle( this.pos, p.pos );
				boolean ckb = checkBelongPeg(p);
				if( ( !ckb ) || blocked( r, p ) ) {
					belongs.remove(i);
				}
			}
			if( belongs.size() == 0 ) {
				throw new ConflictExceptoin("Cell.checkBelong   No peg reaches this cell.");
			}
			if( belongs.size() == 1 ) {				// 所属するペグが１つしかないので【決定】
				// FULLかどうか調べる（ペグの候補矩形で、このセルを含むものが１つだけ（完全確定）かどうか）
				Peg p = belongs.get(0);
				int n = p.countPossibles(this);
				int tp = CELL_UNIQPEG;
				int lv = POINT_CELL_UNIQPEG;
				if( n == 1 ) {
					tp = CELL_UNIQPEG_FULL;
					lv = POINT_CELL_UNIQPEG_FULL;
					// セルの周辺が埋まっている場合
					int	wallcount = countNeighborWall(p.possibles.get(0));
					switch( wallcount ) {
					case 4:
					case 3:
						tp = CELL_UNIQPEG_PLUG;
						lv = POINT_CELL_UNIQPEG_SP1A;
						break;
					case 2:
						tp = CELL_UNIQPEG_PLUG;
						lv = POINT_CELL_UNIQPEG_SP1B;
					} 
				} else if( n == 2 ) {
					lv = POINT_CELL_UNIQPEG_SP2;
				}
				if( (deadend ^ (tp==CELL_UNIQPEG_PLUG)) )
					return;

				Rectangle r = Util.makeRectangle(p.pos, pos);
				bidForNext( new Fixable( tp, lv, this, r ) );		// type, level
			}
		}

		private int	countNeighborWall( Rectangle r ) {
			int	cnt = 0;
			for( int di=0; di<direction.length; ++di ) {
				Cell cur = this;
				for(;;) {
					cur = cur.getPegRelative(direction[di][0],direction[di][1]);
					if( cur == null ) {	// 外に出た　（壁）
						++cnt;
						break;
					}
					if( r.contains(cur.pos) ) {		// 同じ矩形内なので無視
						continue;
					}
					if( cur.peg != null ) {
						++cnt;
					}
					break;
				}
			}

			return	cnt;
		}

		// 指定ペグと自身でできる矩形の中のCellは、すべて指定ペグに属する
		private void applyBelong( int tp, int lev ) {
			if( belongs == null )
				return;

			peg = belongs.get(0);
			belongs = null;			

			Rectangle r = Util.makeRectangle( this.pos, peg.pos );
			for( int i=0; i<r.width; ++i )
				for( int j=0; j<r.height; ++j ) {
					Cell c = cellboard[r.x+i][r.y+j];
					if( c.peg != peg )
						c.peg = peg;
				}
			peg.checkPossiblesCell( tp, lev, this );

			setLevelRectangle( r, tp, lev );
		}

		private boolean checkBelongPeg( Peg p ) {
			// このセルは、ペグに属するか？
			if( p.finished ) {			// 確定ペグ
				return	p.island.rec.contains(pos);
			}

			if( p.num == 0 ) {			// 数字未定のペグの場合、邪魔されていなければOK
				return	true;					// とりあえず、何でもOKにしておく。/////////////////////
			}

			// ペグの可能な矩形のどれかに属せばOK
			for( Rectangle r : p.possibles ) {
				if( ! r.contains( pos ) )
					continue;
				boolean bkd = false;
				lp:
					for( int i=0; i<r.width; ++i ) {
						for( int j=0; j<r.height; ++j ) {
							int x = r.x + i;
							int y = r.y + j;
							Cell c = cellboard[x][y];
							if( c.peg != null && c.peg != p ) {
								bkd = true;
								break lp;
							}
						}
					}
				if( !bkd )
					return true;
			}
			// どの可能矩形にも属さない
			return	false;
		}

		// 矩形の中に、候補ペグ、自身(Cell)以外に所属が決定したCellが存在するかのチェック
		// true : 邪魔されていた
		private boolean blocked( Rectangle r, Peg p ) {
			for( int i=0; i<r.width; ++i )
				for( int j=0; j<r.height; ++j ) {
					int x = r.x + i;
					int y = r.y + j;
					if( p.pos.x == x && p.pos.y == y )
						continue;
					if( p.pos.x == pos.x && p.pos.y == pos.y )
						continue;
					Cell c = cellboard[x][y];
					if( c.ispeg )						// 別のペグが存在した
						return	true;
					if( c.peg!=null ) {
						if( c.peg != p ) {			// 別のペグに属するCellを発見	
							return	true;
						}
					}
				}

			return	false;
		}

		//--------------------------------------------------------------------------
		//	待ち行列、CELL_COMMONの検査
		//	このセルに届くペグの候補矩形の共通矩形は島になる。
		//--------------------------------------------------------------------------
		private void	checkCommomPossibles() {
			if( belongs == null )
				return;

			Rectangle common = null;
			for( Peg p : belongs ) {
				if( p.num == 0 )
					return;
				if( p.possibles == null )
					continue;

				for( Rectangle rec : p.possibles ) {
					if( ! rec.contains( pos ) )
						continue;
					if( common == null ) {
						common = rec;
					} else {
						common = common.intersection(rec);
					}
				}
			}
			if( common == null )
				return;
			if( common.width * common.height == 1 )
				return;

			bidForNext( new Fixable( CELL_COMMON, POINT_CELL_COMMON, this, common ) );		// type, level
		}

		private void	applyCommomPossibles( int tp, int lev, Rectangle common ) {
			if( island != null ) {		// 既存の島に含まれているならば、無視
				Rectangle andrec = island.rec.intersection(common);
				if( common.equals(andrec) ) {
					return;
				}
			}

			Island	ild = new Island( new Rectangle(pos.x,pos.y,1,1), null );
			ild.expandTo( common );

			setLevelRectangle( common, tp, lev );
		}

		//	相対位置　v　のセルを求める
		private Cell	getPegRelative( int x, int y ) {
			Point p = new Point( pos.x+x, pos.y+y );
			if( ! isOnBoard(p) )
				return	null;
			return	cellboard[p.x][p.y];
		}
	}

	//==============================================================================
	//	決定可能場所
	//==============================================================================
	class	Fixable {
		int		type;
		int		level;
		Peg		peg;
		Cell	cell;
		Rectangle	rec;
		double	diff;			// 難易度ポイント			// とりあえず　level

		Fixable( int t, int l, Peg p, Rectangle r ) {
			type = t;
			level = l;
			peg = p;
			rec = r;
			diff = level;
			peg.wait = true;
		}		

		Fixable( int t, int l, Cell c, Rectangle r ) {
			type = t;
			level = l;
			cell = c;
			rec = r;
			diff = level;		
		}
		Fixable( int t, int l, Rectangle r ) {
			type = t;
			level = l;
			rec = r;
			diff = level;
		}
	}

	void	bidForNext( Fixable f ) {
		if( nextfixable == null ) {
			nextfixable = f;
		} else if( f.level < nextfixable.level ) {
			nextfixable = f;
		}

		currentdiffpoint += 1.0 / f.level;
	}

	Fixable	getCurrentStep() {
		return	answersequence.get(answersequence.size()-1);
	}

	int	getDiffPoint() {
		return	(int)totaldiffpoint;
	}

	//==========================================================================
	//		領域分割
	//==========================================================================
	class Region {
		boolean[][] orgfix;		//
		boolean[][] wrkfix;
		int	[][]	orgregion;		// 現盤面のアキマス部分を領域分けしたもの
		int	[][]	wrkregion;		// 	作業用領域
		int	regcount;				// 領域数
		int minpegcnterr;			// エラー時の最小ペグ数

		Region() {
			getNonfixArea();
			orgregion = findRegion(orgfix);
		}

		private boolean checkRectangle( Rectangle rec ) {
			getWrkfix(rec);
			wrkregion = findRegion(wrkfix);
			int [] regsizes = new int[regcount+1];
			int	[]	regnumsum = new int[regcount+1];
			int []	regpegcnt = new int[regcount+1];
			boolean[] incunknown = new boolean[regcount+1];
			for( int i=0; i<width; ++i )
				for( int j=0; j<height; ++j ) {
					int	rnum = wrkregion[i][j];
					if( rnum > 0 ) {
						++regsizes[rnum];
					}
					Cell c = cellboard[i][j];
					if( c.ispeg ) {
						Peg p = c.peg;
						if( p.num == 0 ) {
							incunknown[rnum] = true;
						} else {
							regnumsum[rnum] += p.num;
							++regpegcnt[rnum];
						}
					}
				}

			// 分割不一致検査
			boolean	possible = true;
			minpegcnterr = 1000000;
			for( int rn=1; rn<=regcount; ++rn ) {
				if( regsizes[rn]!=regnumsum[rn] && !incunknown[rn] ) {
					possible = false;
					if( minpegcnterr > regpegcnt[rn] )
						minpegcnterr = regpegcnt[rn];
				}
			}

			return	possible;
		}

		// 	初期設定
		private void	getNonfixArea() {
			int[][] bd = getCellIds();
			orgfix = new boolean[width][height];
			for( int i=0; i<width; ++i )
				for( int j=0; j<height; ++j ) {
					orgfix[i][j] = bd[i][j] <= 0;
				}
		}

		// 領域分割
		private int[][] findRegion(boolean[][] fix) {
			int[][] reg = new int[width][height];

			int	regnum = 0;
			for( int i=0; i<width; ++i )
				for( int j=0; j<height; ++j ) {
					if( ! fix[i][j] )
						continue;
					if( reg[i][j] != 0 )
						continue;
					++regnum;
					Stack<Point> wt = new Stack<Point>();
					wt.push(new Point(i,j));
					while( wt.size() > 0 ) {
						Point pos = wt.pop();
						if( pos.x <0 || pos.x>=width || pos.y<0 || pos.y>=height )
							continue;
						if( !fix[pos.x][pos.y] )
							continue;
						if( reg[pos.x][pos.y] != 0 )
							continue;
						reg[pos.x][pos.y] = regnum;
						for( int k=0; k<direction.length; ++k ) {
							wt.push( new Point(pos.x+direction[k][0],pos.y+direction[k][1]));
						}
					}
				}

			regcount = regnum;
			return	reg;
		}

		void getWrkfix(Rectangle rec) {
			wrkfix = new boolean[width][height];
			for(int i=0; i<width; ++i)
				for( int j=0; j<height; ++j )
					wrkfix[i][j] = orgfix[i][j];

			for(int i=0; i<rec.width; ++i)
				for( int j=0; j<rec.height; ++j )
					wrkfix[rec.x+i][rec.y+j] = false;
		}
	}
}

class	Difficulty {
	public		boolean	deadend;
	public		boolean	initfixhalf;
	public		boolean	pegfixhalf;
	public		boolean	celluniqpeg;
	public		boolean	cellcommon;
	public		boolean	regiondivide;

	Difficulty() {
		deadend = initfixhalf = pegfixhalf = celluniqpeg = cellcommon = true;
		regiondivide = false;
	}

	Difficulty(boolean b) {
		deadend = initfixhalf = pegfixhalf = celluniqpeg = cellcommon = regiondivide = b;
	}
}
